from __future__ import annotations

import types
import typing as t

import responses

from globus_sdk._internal.utils import slash_join


class RegisteredResponse:
    """
    A mock response along with descriptive metadata to let a fixture "pass data
    forward" to the consuming test cases. (e.g. a ``GET Task`` fixture which
    shares the ``task_id`` it uses with consumers via ``.metadata["task_id"]``)

    When initializing a ``RegisteredResponse``, you can use ``path`` and ``service``
    to describe a path on a Globus service rather than a full URL. The ``metadata``
    data container is also globus-sdk specific. Most other parameters are wrappers
    over ``responses`` response characteristics.

    :param path: Path on the target service or full URL if service is null
    :param service: A known service name like ``"transfer"`` or ``"compute"``. This will
        be used to deduce the base URL onto which ``path`` should be joined
    :param method: A string HTTP Method
    :param headers: HTTP headers for the response
    :param json: A dict or list structure for a JSON response (mutex with ``body``)
    :param body: A string response body (mutex with ``json``)
    :param status: The HTTP status code for the response
    :param content_type: A Content-Type header value for the response
    :param match: A tuple or list of ``responses`` matchers
    :param metadata: A dict of data to store on the response, which allows the usage
        site which declares the response to pass information forward to the site which
        activates and tests against the response.
    """

    _url_map = {
        "auth": "https://auth.globus.org/",
        "nexus": "https://nexus.api.globusonline.org/",
        "transfer": "https://transfer.api.globus.org/",
        "search": "https://search.api.globus.org/",
        "gcs": "https://abc.xyz.data.globus.org/api/",
        "groups": "https://groups.api.globus.org/",
        "timer": "https://timer.automate.globus.org/",
        "flows": "https://flows.automate.globus.org/",
        "compute": "https://compute.api.globus.org/",
    }

    def __init__(
        self,
        *,
        # path and service are glbous-sdk specific
        # in `responses`, these are just `url`
        path: str,
        service: (
            t.Literal[
                "auth",
                "nexus",
                "transfer",
                "search",
                "gcs",
                "groups",
                "timer",
                "flows",
                "compute",
            ]
            | None
        ) = None,
        # method will be passed through to `responses.Response`, so we
        # support all of the values which it supports
        method: t.Literal[
            "GET",
            "PUT",
            "POST",
            "PATCH",
            "HEAD",
            "DELETE",
            "OPTIONS",
            "CONNECT",
            "TRACE",
        ] = "GET",
        # these parameters are passed through to `response.Response` (or omitted)
        body: str | None = None,
        content_type: str | None = None,
        headers: dict[str, str] | None = None,
        json: None | list[t.Any] | dict[str, t.Any] = None,
        status: int = 200,
        stream: bool | None = None,
        match: t.Sequence[t.Callable[..., tuple[bool, str]]] | None = None,
        # metadata is globus-sdk specific
        metadata: dict[str, t.Any] | None = None,
        # the following are known parameters to `responses.Response` which
        # `RegisteredResponse` does not support:
        #   - url: calculated from (path, service)
        #   - auto_calculate_content_length: a bool setting, usually not needed and can
        #                                    be achieved in user code via `headers`
        #   - passthrough: bool setting allowing calls to be emitted to the services
        #                  (undesirable in any ordinary cases)
        #   - match_querystring: legacy param which has been replaced with `match`
    ) -> None:
        self.service = service

        if service:
            self.full_url = slash_join(self._url_map[service], path)
        else:
            self.full_url = path

        self.path = path

        # convert the method to uppercase so that specifying `method="post"` will match
        # correctly -- method matching is case sensitive but we don't need to expose the
        # possibility of a non-uppercase method
        self.method = method.upper()

        self.body = body
        self.content_type = content_type
        self.headers = headers
        self.json = json
        self.status = status
        self.stream = stream

        self.match = match

        self._metadata = metadata

        self.parent: ResponseSet | ResponseList | None = None

    @property
    def metadata(self) -> dict[str, t.Any]:
        if self._metadata is not None:
            return self._metadata
        if self.parent is not None:
            return self.parent.metadata
        return {}

    def _add_or_replace(
        self,
        method: t.Literal["add", "replace"],
        *,
        requests_mock: responses.RequestsMock | None = None,
    ) -> RegisteredResponse:
        kwargs: dict[str, t.Any] = {
            "headers": self.headers,
            "status": self.status,
            "stream": self.stream,
            "match_querystring": None,
        }
        if self.json is not None:
            kwargs["json"] = self.json
        if self.body is not None:
            kwargs["body"] = self.body
        if self.content_type is not None:
            kwargs["content_type"] = self.content_type
        if self.match is not None:
            kwargs["match"] = self.match

        if requests_mock is None:
            use_requests_mock: responses.RequestsMock | types.ModuleType = responses
        else:
            use_requests_mock = requests_mock

        if method == "add":
            use_requests_mock.add(self.method, self.full_url, **kwargs)
        else:
            use_requests_mock.replace(self.method, self.full_url, **kwargs)
        return self

    def add(
        self, *, requests_mock: responses.RequestsMock | None = None
    ) -> RegisteredResponse:
        """
        Activate the response, adding it to a mocked requests object.

        :param requests_mock: The mocked requests object to use. Defaults to the default
            provided by the ``responses`` library
        """
        return self._add_or_replace("add", requests_mock=requests_mock)

    def replace(
        self, *, requests_mock: responses.RequestsMock | None = None
    ) -> RegisteredResponse:
        """
        Activate the response, adding it to a mocked requests object and replacing any
        existing response for the particular path and method.

        :param requests_mock: The mocked requests object to use. Defaults to the default
            provided by the ``responses`` library
        """
        return self._add_or_replace("replace", requests_mock=requests_mock)


class ResponseList:
    """
    A series of unnamed responses, meant to be used and referred to as a single case
    within a ResponseSet.

    This can be stored in a ``ResponseSet`` as a case, describing a series
    of responses registered to a specific name (e.g. to describe a paginated API).
    """

    def __init__(
        self,
        *data: RegisteredResponse,
        metadata: dict[str, t.Any] | None = None,
    ) -> None:
        self.responses = list(data)
        self._metadata = metadata
        self.parent: ResponseSet | None = None
        for r in data:
            r.parent = self

    @property
    def metadata(self) -> dict[str, t.Any]:
        if self._metadata is not None:
            return self._metadata
        if self.parent is not None:
            return self.parent.metadata
        return {}

    def add(
        self, *, requests_mock: responses.RequestsMock | None = None
    ) -> ResponseList:
        for r in self.responses:
            r.add(requests_mock=requests_mock)
        return self


class ResponseSet:
    """
    A collection of mock responses, potentially all meant to be activated together
    (``.activate_all()``), or to be individually selected as options/alternatives
    (``.activate("case_foo")``).

    On init, this implicitly sets the parent of any response objects to this response
    set. On register() it does not do so automatically.
    """

    def __init__(
        self,
        metadata: dict[str, t.Any] | None = None,
        **kwargs: RegisteredResponse | ResponseList,
    ) -> None:
        self.metadata = metadata or {}
        self._data: dict[str, RegisteredResponse | ResponseList] = {**kwargs}
        for res in self._data.values():
            res.parent = self

    def register(self, case: str, value: RegisteredResponse) -> None:
        self._data[case] = value

    def lookup(self, case: str) -> RegisteredResponse | ResponseList:
        try:
            return self._data[case]
        except KeyError as e:
            raise LookupError("did not find a matching registered response") from e

    def __bool__(self) -> bool:
        return bool(self._data)

    def __iter__(
        self,
    ) -> t.Iterator[RegisteredResponse | ResponseList]:
        return iter(self._data.values())

    def cases(self) -> t.Iterator[str]:
        return iter(self._data)

    def activate(
        self,
        case: str,
        *,
        requests_mock: responses.RequestsMock | None = None,
    ) -> RegisteredResponse | ResponseList:
        return self.lookup(case).add(requests_mock=requests_mock)

    def activate_all(
        self, *, requests_mock: responses.RequestsMock | None = None
    ) -> ResponseSet:
        for x in self:
            x.add(requests_mock=requests_mock)
        return self

    @classmethod
    def from_dict(
        cls,
        data: t.Mapping[
            str,
            (dict[str, t.Any] | list[dict[str, t.Any]]),
        ],
        metadata: dict[str, t.Any] | None = None,
        **kwargs: dict[str, dict[str, t.Any]],
    ) -> ResponseSet:
        # constructor which expects native dicts and converts them to RegisteredResponse
        # objects, then puts them into the ResponseSet
        def handle_value(
            v: dict[str, t.Any] | list[dict[str, t.Any]],
        ) -> RegisteredResponse | ResponseList:
            if isinstance(v, dict):
                return RegisteredResponse(**v)
            else:
                return ResponseList(*(RegisteredResponse(**subv) for subv in v))

        reassembled_data: dict[str, RegisteredResponse | ResponseList] = {
            k: handle_value(v) for k, v in data.items()
        }
        return cls(metadata=metadata, **reassembled_data)



import globus_sdk
from globus_sdk.testing import load_response


def test_get_endpoint_status(compute_client_v2: globus_sdk.ComputeClientV2):
    meta = load_response(compute_client_v2.get_endpoint_status).metadata
    res = compute_client_v2.get_endpoint_status(endpoint_id=meta["endpoint_id"])
    assert res.http_status == 200
    assert res.data["status"] == "online"



from globus_sdk import TimersAPIError
from globus_sdk.testing import construct_error


def test_timer_error_load_simple():
    err = construct_error(
        error_class=TimersAPIError,
        body={"error": {"code": "ERROR", "detail": "Request failed", "status": 500}},
        http_status=500,
    )

    assert err.code == "ERROR"
    assert err.message == "Request failed"


def test_timer_error_load_nested():
    err = construct_error(
        error_class=TimersAPIError,
        body={
            "detail": [
                {
                    "loc": ["body", "start"],
                    "msg": "field required",
                    "type": "value_error.missing",
                },
                {
                    "loc": ["body", "end"],
                    "msg": "field required",
                    "type": "value_error.missing",
                },
            ]
        },
        http_status=422,
    )

    assert err.code is None
    assert err.message == "field required: body.start; field required: body.end"


def test_timer_error_load_unrecognized_format():
    err = construct_error(error_class=TimersAPIError, body={}, http_status=400)
    assert err.code is None
    assert err.message is None



import pytest

import globus_sdk
from globus_sdk.testing import load_response


def test_set_subscription_admin_verified(client):
    meta = load_response(client.set_subscription_admin_verified).metadata
    epid = meta["endpoint_id"]

    res = client.set_subscription_admin_verified(epid, True)
    assert res["code"] == "Updated"
    assert res["message"] == "Endpoint updated successfully"


def test_set_subscription_admin_verified_fails_no_admin_role(client):
    meta = load_response(
        client.set_subscription_admin_verified, case="no_admin_role"
    ).metadata
    epid = meta["endpoint_id"]

    with pytest.raises(globus_sdk.TransferAPIError) as excinfo:
        client.set_subscription_admin_verified(epid, True)

    assert excinfo.value.code == "PermissionDenied"
    assert len(excinfo.value.messages) == 1
    message = excinfo.value.messages[0]
    assert message == (
        "User does not have an admin role on the collection's subscription to set "
        "subscription_admin_verified"
    )


def test_set_subscription_admin_verified_fails_non_valid_verified_status(client):
    meta = load_response(
        client.set_subscription_admin_verified, case="non_valid_verified_status"
    ).metadata
    epid = meta["endpoint_id"]

    with pytest.raises(globus_sdk.TransferAPIError) as excinfo:
        client.set_subscription_admin_verified(epid, None)

    assert excinfo.value.code == "BadRequest"
    assert len(excinfo.value.messages) == 1
    message = excinfo.value.messages[0]
    assert message.startswith("Could not parse JSON: ")


def test_set_subscription_admin_verified_fails_non_subscribed_endpoint(client):
    meta = load_response(
        client.set_subscription_admin_verified, case="non_subscribed_endpoint"
    ).metadata
    epid = meta["endpoint_id"]

    with pytest.raises(globus_sdk.TransferAPIError) as excinfo:
        client.set_subscription_admin_verified(epid, True)

    assert excinfo.value.code == "BadRequest"
    assert len(excinfo.value.messages) == 1
    message = excinfo.value.messages[0]
    assert message == (
        "The collection must be associated with a subscription to "
        "set subscription_admin_verified"
    )


def test_set_subscription_admin_verified_fails_no_identities_in_session(client):
    meta = load_response(
        client.set_subscription_admin_verified, case="no_identities_in_session"
    ).metadata
    epid = meta["endpoint_id"]
    subid = meta["subscription_id"]

    with pytest.raises(globus_sdk.TransferAPIError) as excinfo:
        client.set_subscription_admin_verified(epid, True)

    assert excinfo.value.code == "BadRequest"
    assert len(excinfo.value.messages) == 1
    message = excinfo.value.messages[0]
    assert message == (
        "No manager or admin identities in session for high-assurance subscription "
        f"{subid}"
    )



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import ENDPOINT_ID, FUNCTION_ID, FUNCTION_ID_2

DEFAULT_RESPONSE_DOC = {
    "endpoint_id": ENDPOINT_ID,
    "restricted": True,
    "functions": [FUNCTION_ID, FUNCTION_ID_2],
}

RESPONSES = ResponseSet(
    metadata={"endpoint_id": ENDPOINT_ID},
    default=RegisteredResponse(
        service="compute",
        path=f"/v3/endpoints/{ENDPOINT_ID}/allowed_functions",
        method="GET",
        json=DEFAULT_RESPONSE_DOC,
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TIMER_ID, V1_TIMER

RESPONSES = ResponseSet(
    metadata={"job_id": TIMER_ID},
    default=RegisteredResponse(
        service="timer",
        path=f"/jobs/{TIMER_ID}",
        method="DELETE",
        json=V1_TIMER,
    ),
)



"""
Test that various modules do not rely on `requests` at import-time.

This means that these modules are safe to import in highly latency-sensitive
applications like globus-cli.
"""

import os
import subprocess
import sys

import pytest

PYTHON_BINARY = os.environ.get("GLOBUS_TEST_PY", sys.executable)


@pytest.mark.parametrize(
    "module_name",
    (
        # most of the SDK should not pull in 'requests', making the parts which do
        # not handle request sending easy to use without the perf penalty from
        # requests/urllib3
        "authorizers",
        "config",
        "gare",
        "local_endpoint",
        "login_flows",
        "paging",
        "response",
        "scopes",
        "token_storage",
        # the top-level of the 'exc' subpackage (but not necessarily its contents)
        # should similarly be standalone, for exception handlers
        "exc",
        # internal components and utilities are a special case:
        # failing to ensure that these avoid 'requests' can make it more difficult
        # to ensure that the main parts (above) do not transitively pick it up
        "_internal.classprop",
        "_internal.guards",
        "_internal.remarshal",
        "_internal.serializable",
        "_internal.utils",
        "_internal.type_definitions",
        "_missing",
    ),
)
def test_module_does_not_require_requests(module_name):
    command = (
        f"import globus_sdk.{module_name}; "
        "import sys; assert 'requests' not in sys.modules"
    )
    proc = subprocess.Popen(
        f'{PYTHON_BINARY} -c "{command}"',
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    status = proc.wait()
    assert status == 0, str(proc.communicate())
    proc.stdout.close()
    proc.stderr.close()



import dataclasses
import urllib.parse

import pytest

from globus_sdk._missing import MISSING
from globus_sdk.testing import get_last_request, load_response


def test_search_index_list(client):
    meta = load_response(client.index_list).metadata
    index_ids = meta["index_ids"]

    res = client.index_list()
    assert res.http_status == 200

    index_list = res["index_list"]
    assert isinstance(index_list, list)
    assert len(index_list) == len(index_ids)
    assert [i["id"] for i in index_list] == index_ids


def test_search_index_list_is_iterable(client):
    meta = load_response(client.index_list).metadata
    index_ids = meta["index_ids"]

    res = client.index_list()
    assert res.http_status == 200

    index_list = list(res)
    assert len(index_list) == len(index_ids)
    assert [i["id"] for i in index_list] == index_ids


_OMIT = object()


@dataclasses.dataclass
class FilterRoleParam:
    name: str
    value: object
    expect_parsed_param: object = None

    @property
    def missing(self) -> bool:
        return self.value in (_OMIT, MISSING)

    @property
    def call_kwargs(self) -> dict[str, object]:
        return {} if self.value is _OMIT else {"filter_roles": self.value}

    def __str__(self) -> str:
        return self.name


@pytest.mark.parametrize(
    "filter_param",
    [
        FilterRoleParam("omitted", _OMIT),
        FilterRoleParam("missing", MISSING),
        FilterRoleParam("ownerstr", "owner", ["owner"]),
        FilterRoleParam("adminstr", "admin", ["admin"]),
        FilterRoleParam("writerstr", "writer", ["writer"]),
        FilterRoleParam("tuple", ("owner", "admin"), ["owner,admin"]),
        FilterRoleParam("list", ["admin", "writer"], ["admin,writer"]),
        FilterRoleParam("duplicates", ("admin", "admin"), ["admin,admin"]),
        # this isn't a real role, but it should be passed through as though it were
        # -- this value can be a typing time error but never a runtime error
        FilterRoleParam("unknown_str", "ambassador", ["ambassador"]),
    ],
    ids=str,
)
def test_search_index_list_encodes_filter_roles_as_expected(client, filter_param):
    load_response(client.index_list).metadata
    res = client.index_list(**filter_param.call_kwargs)
    assert res.http_status == 200

    req = get_last_request()
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    if filter_param.missing:
        assert "filter_roles" not in parsed_qs
    else:
        assert parsed_qs["filter_roles"] == filter_param.expect_parsed_param



from __future__ import annotations

import datetime
import typing as t
import uuid

from responses import matchers

from globus_sdk.testing.models import RegisteredResponse, ResponseList, ResponseSet

from ._common import TWO_HOP_TRANSFER_FLOW_DOC, TWO_HOP_TRANSFER_FLOW_ID

OWNER_ID = "e061df5a-b7b9-4578-a73b-6d4a4edfd66e"


def generate_hello_world_example_flow(n: int) -> dict[str, t.Any]:
    flow_id = str(uuid.UUID(int=n))
    base_time = datetime.datetime.fromisoformat("2021-10-18T19:19:35.967289+00:00")
    updated_at = created_at = base_time + datetime.timedelta(days=n)
    flow_user_scope = (
        f"https://auth.globus.org/scopes/{flow_id}/"
        f"flow_{flow_id.replace('-', '_')}_user"
    )

    return {
        "action_url": f"https://flows.automate.globus.org/flows/{flow_id}",
        "administered_by": [],
        "api_version": "1.0",
        "created_at": created_at.isoformat() + "+00:00",
        "created_by": f"urn:globus:auth:identity:{OWNER_ID}",
        "definition": {
            "StartAt": "HelloWorld",
            "States": {
                "HelloWorld": {
                    "ActionScope": (
                        "https://auth.globus.org/scopes/actions.globus.org/hello_world"
                    ),
                    "ActionUrl": "https://actions.globus.org/hello_world",
                    "End": True,
                    "Parameters": {"echo_string": "Hello, World."},
                    "ResultPath": "$.Result",
                    "Type": "Action",
                }
            },
        },
        "description": "A simple Flow...",
        "flow_administrators": [],
        "flow_owner": f"urn:globus:auth:identity:{OWNER_ID}",
        "flow_starters": [],
        "flow_url": f"https://flows.automate.globus.org/flows/{flow_id}",
        "flow_viewers": [],
        "globus_auth_scope": flow_user_scope,
        "globus_auth_username": f"{flow_id}@clients.auth.globus.org",
        "id": str(flow_id),
        "input_schema": {
            "additionalProperties": False,
            "properties": {
                "echo_string": {"description": "The string to echo", "type": "string"},
                "sleep_time": {"type": "integer"},
            },
            "required": ["echo_string", "sleep_time"],
            "type": "object",
        },
        "keywords": [],
        "log_supported": True,
        "principal_urn": f"urn:globus:auth:identity:{flow_id}",
        "runnable_by": [],
        "subtitle": "",
        "synchronous": False,
        "title": f"Hello, World (Example {n})",
        "types": ["Action"],
        "updated_at": updated_at.isoformat() + "+00:00",
        "user_role": "flow_viewer",
        "visible_to": [],
    }


RESPONSES = ResponseSet(
    metadata={"first_flow_id": TWO_HOP_TRANSFER_FLOW_ID},
    default=RegisteredResponse(
        service="flows",
        path="/flows",
        json={
            "flows": [TWO_HOP_TRANSFER_FLOW_DOC],
            "limit": 20,
            "has_next_page": False,
            "marker": None,
        },
    ),
    paginated=ResponseList(
        RegisteredResponse(
            service="flows",
            path="/flows",
            json={
                "flows": [generate_hello_world_example_flow(i) for i in range(20)],
                "limit": 20,
                "has_next_page": True,
                "marker": "fake_marker_0",
            },
        ),
        RegisteredResponse(
            service="flows",
            path="/flows",
            json={
                "flows": [generate_hello_world_example_flow(i) for i in range(20, 40)],
                "limit": 20,
                "has_next_page": True,
                "marker": "fake_marker_1",
            },
            match=[matchers.query_param_matcher({"marker": "fake_marker_0"})],
        ),
        RegisteredResponse(
            service="flows",
            path="/flows",
            json={
                "flows": [generate_hello_world_example_flow(i) for i in range(40, 60)],
                "limit": 20,
                "has_next_page": False,
                "marker": None,
            },
            match=[matchers.query_param_matcher({"marker": "fake_marker_1"})],
        ),
        metadata={
            "owner_id": OWNER_ID,
            "num_pages": 3,
            "expect_markers": ["fake_marker_0", "fake_marker_1", None],
            "total_items": 60,
        },
    ),
)



from __future__ import annotations

import typing as t
import uuid

if t.TYPE_CHECKING:
    from globus_sdk import AuthLoginClient, IDTokenDecoder
    from globus_sdk.login_flows import LoginFlowManager
    from globus_sdk.token_storage import TokenStorage, TokenValidationError

    from .app import GlobusApp
    from .config import GlobusAppConfig


@t.runtime_checkable
class TokenStorageProvider(t.Protocol):
    r"""
    A protocol for a factory which can create ``TokenStorage``\s.

    SDK-provided :ref:`token_storages` support this protocol.
    """

    @classmethod
    def for_globus_app(
        cls,
        *,
        app_name: str,
        config: GlobusAppConfig,
        client_id: uuid.UUID | str,
        namespace: str,
    ) -> TokenStorage:
        """
        Create a ``TokenStorage`` for use in a GlobusApp.

        :param app_name: The name supplied to the GlobusApp.
        :param config: The configuration supplied to the GlobusApp.
        :param client_id: The client_id of to the GlobusApp.
        :param namespace: A namespace to use for instantiating a TokenStorage.
        """


@t.runtime_checkable
class LoginFlowManagerProvider(t.Protocol):
    r"""
    A protocol for a factory which can create ``LoginFlowManager``\s.

    SDK-provided :ref:`login_flow_managers` support this protocol.
    """

    @classmethod
    def for_globus_app(
        cls, *, app_name: str, config: GlobusAppConfig, login_client: AuthLoginClient
    ) -> LoginFlowManager:
        """
        Create a ``CommandLineLoginFlowManager`` for use in a GlobusApp.

        :param app_name: The name supplied to the GlobusApp.
        :param config: The configuration supplied to the GlobusApp.
        :param login_client: A login client to use for instantiating a LoginFlowManager.
        """


@t.runtime_checkable
class IDTokenDecoderProvider(t.Protocol):
    r"""
    A protocol for a factory which can create ``IDTokenDecoder``\s.

    The SDK-provided ``IDTokenDecoder`` class supports this protocol.
    """

    @classmethod
    def for_globus_app(
        cls,
        *,
        app_name: str,
        config: GlobusAppConfig,
        login_client: AuthLoginClient,
    ) -> IDTokenDecoder:
        """
        Create an ``IDTokenDecoder`` for use in a GlobusApp.

        :param app_name: The name supplied to the GlobusApp.
        :param config: The configuration supplied to the GlobusApp.
        :param login_client: A login client to use for instantiating an
            ``IDTokenDecoder``.
        """


class TokenValidationErrorHandler(t.Protocol):
    """
    A handler invoked when a :class:`TokenValidationError` is raised during a
    service client call.

    If this call completes without raising an exception, the service client call
    will retry once more (subsequent errors will not call this handler).
    """

    def __call__(self, app: GlobusApp, error: TokenValidationError) -> None:
        """
        Resolve a token validation error.

        :param app: The GlobusApp instance which encountered an error.
        :param error: The error which was encountered.
        """



from types import MappingProxyType

from globus_sdk import ClientApp, UserApp

my_user_app = UserApp("...", client_id="...")
my_client_app = ClientApp("...", client_id="...", client_secret="...")

# declare scope data in the form of a subtype of the
# `str | Scope | t.Iterable[str | Scope]` (`list[str]`) indexed in a dict,
# this is meant to be a subtype of the requirements data accepted by
# `GlobusApp.add_scope_requirements`
#
# this is a regression test for that being annotated as
# `dict[str, str | Scope | t.Iterable[str | Scope]]` which will reject the input
# type because `dict` is a mutable container, and therefore invariant
scopes: dict[str, list[str]] = {"foo": ["bar"]}
my_user_app.add_scope_requirements(scopes)
my_client_app.add_scope_requirements(scopes)

# a mapping proxy is an immutable mapping (proxy) and should be accepted by apps as well
# meaning that any mapping is fine, not just `dict` specifically (or MutableMapping)
my_user_app.add_scope_requirements(MappingProxyType(scopes))
my_client_app.add_scope_requirements(MappingProxyType(scopes))


# both of the above tests repeated, but now on init
my_user_app = UserApp("...", client_id="...", scope_requirements=scopes)
my_user_app = UserApp(
    "...", client_id="...", scope_requirements=MappingProxyType(scopes)
)
my_client_app = ClientApp(
    "...", client_id="...", client_secret="...", scope_requirements=scopes
)
my_client_app = ClientApp(
    "...",
    client_id="...",
    client_secret="...",
    scope_requirements=MappingProxyType(scopes),
)



import importlib.metadata
import logging
import sys

from ._internal.lazy_import import (
    default_dir_implementation,
    default_getattr_implementation,
    load_all_tuple,
)

__version__ = importlib.metadata.distribution("globus_sdk").version


def _force_eager_imports() -> None:
    current_module = sys.modules[__name__]

    for attr in __all__:
        getattr(current_module, attr)


#
# all lazy SDK attributes are defined in __init__.pyi
#
# to add an attribute, write the relevant import in `__init__.pyi` and update
# the `__all__` tuple there
#
__all__ = load_all_tuple(__name__, "__init__.pyi")
__getattr__ = default_getattr_implementation(__name__, "__init__.pyi")
__dir__ = default_dir_implementation(__name__)

del load_all_tuple
del default_getattr_implementation
del default_dir_implementation


# configure logging for a library, per python best practices:
# https://docs.python.org/3/howto/logging.html#configuring-logging-for-a-library
logging.getLogger("globus_sdk").addHandler(logging.NullHandler())



import uuid

import pytest

from globus_sdk.testing import get_last_request, load_response


@pytest.mark.parametrize("use_uuid", [False, True])
def test_get_registered_api(flows_client, use_uuid):
    """Test get_registered_api with string ID and UUID ID"""
    loaded_response = load_response(flows_client.get_registered_api)
    _, meta = loaded_response.json, loaded_response.metadata

    registered_api_id_str = meta["registered_api_id"]
    registered_api_id = (
        uuid.UUID(registered_api_id_str) if use_uuid else registered_api_id_str
    )

    res = flows_client.get_registered_api(registered_api_id)

    assert res.http_status == 200
    assert res["id"] == registered_api_id_str
    assert res["name"] == meta["name"]

    req = get_last_request()
    assert req.body is None
    assert f"/registered_apis/{registered_api_id_str}" in req.url



from __future__ import annotations

import typing as t

from globus_sdk import exc

from ._graph_parser import ScopeGraph
from .representation import Scope


class ScopeParser:
    """
    The ``ScopeParser`` handles the conversion of strings to scopes.

    Most interfaces are classmethods, meaning users should prefer usage like
    ``ScopeParser.parse("foo")``
    """

    @classmethod
    def parse(cls, scope_string: str) -> list[Scope]:
        """
        Parse an arbitrary scope string to a list of scopes.

        Zero or more than one scope may be returned, as in the case of an empty string
        or space-delimited scopes.

        .. warning::

            Parsing passes through an intermediary representation which treats scopes
            as a graph. This ensures that the behavior of parses matches the treatment
            of scope strings in Globus Auth authorization flows.
            However, this also means that the parsing does not allow for strings which
            represent consent trees with structures in which the same scope appears in
            multiple parts of the tree.

        :param scope_string: The string to parse
        """
        # build the graph intermediate representation
        scope_graph = ScopeGraph.parse(scope_string)

        # traverse the graph in a reversed BFS scan
        #
        # this means we'll handle leaf nodes first, and we'll never reach a
        # node before its descendants (dependencies)
        #
        # as we work, build a lookup table for built Scope objects so that we can
        # quickly retrieve elements
        built_scopes: dict[tuple[str, bool], Scope] = {}

        for name, optionality in list(scope_graph.breadth_first_walk())[::-1]:
            dependencies: tuple[Scope, ...] = tuple(
                # the lookup in built_scopes here is safe because of the
                # reversed BFS ordering
                built_scopes[(dep_name, dep_optional)]
                for _, dep_name, dep_optional in scope_graph.adjacency_matrix[name]
            )

            built_scopes[(name, optionality)] = Scope(
                name, optional=optionality, dependencies=dependencies
            )

        # only return the top-level elements from that build process
        # (the roots of the forest-shaped graph)
        return [built_scopes[key] for key in scope_graph.top_level_scopes]

    @classmethod
    def merge_scopes(cls, scopes_a: list[Scope], scopes_b: list[Scope]) -> list[Scope]:
        """
        Given two lists of Scopes, merge them into one list of Scopes by parsing
        them as one combined scope string.

        :param scopes_a: list of Scopes to be merged with scopes_b
        :param scopes_b: list of Scopes to be merged with scopes_a
        """
        # dict of base scope_string: list of scopes with that base scope_string
        return cls.parse(
            " ".join([str(s) for s in scopes_a] + [str(s) for s in scopes_b])
        )

    @classmethod
    def serialize(
        cls, scopes: str | Scope | t.Iterable[str | Scope], *, reject_empty: bool = True
    ) -> str:
        """
        Normalize scopes to a space-separated scope string.

        The results of this method are suitable for sending to Globus Auth.
        Scopes are not parsed, merged, or normalized by this method.

        :param scopes: A scope string, scope object, or an iterable of scope strings
            and scope objects.
        :param reject_empty: When true (the default), raise an error if the
            scopes serialize to the empty string.
        :returns: A space-separated scope string.

        Example usage:

        .. code-block:: pycon

            >>> ScopeParser.serialize([Scope("foo"), "bar", Scope("qux")])
            'foo bar qux'
        """
        scope_iter: t.Iterable[str | Scope]
        if isinstance(scopes, (str, Scope)):
            scope_iter = (scopes,)
        else:
            scope_iter = scopes

        result = " ".join(str(scope) for scope in scope_iter)
        if reject_empty and result == "":
            raise exc.GlobusSDKUsageError(
                "'scopes' cannot be the empty string or empty collection."
            )
        return result



import globus_sdk
from globus_sdk.services.auth._common import SupportsJWKMethods

# setup clients
nc = globus_sdk.NativeAppAuthClient("foo_client_id")
cc = globus_sdk.ConfidentialAppAuthClient("foo_client_id", "foo_client_secret")

# check that each one supports the JWK methods
x: SupportsJWKMethods
x = nc
x = cc



import uuid

from globus_sdk.testing import RegisteredResponse, ResponseList, ResponseSet

from ._common import RUN_ID

GET_RUN_DEFINITION = {
    "flow_id": str(uuid.uuid4()),
    "definition": {
        "States": {"no-op": {"End": True, "Type": "Pass"}},
        "StartAt": "no-op",
    },
    "input_schema": {},
}


RESPONSES = ResponseSet(
    metadata={"run_id": RUN_ID},
    default=ResponseList(
        RegisteredResponse(
            service="flows",
            method="GET",
            path=f"/runs/{RUN_ID}/definition",
            json=GET_RUN_DEFINITION,
        ),
    ),
)



from __future__ import annotations

import uuid

import pytest

from globus_sdk.testing import load_response


@pytest.mark.parametrize(
    "uuid_type",
    (str, uuid.UUID),
)
def test_delete_policy(
    service_client,
    uuid_type: type[str] | type[uuid.UUID],
):
    meta = load_response(service_client.delete_policy).metadata

    if uuid_type is str:
        res = service_client.delete_policy(meta["policy_id"])
    else:
        res = service_client.delete_policy(uuid.UUID(meta["policy_id"]))

    assert res["policy"]["id"] == meta["policy_id"]



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import (
    DEST_EP_ID,
    FLOW_ID,
    SOURCE_EP_ID,
    TIMER_ID,
    V2_FLOW_TIMER,
    V2_TRANSFER_TIMER,
)

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="timer",
        path="/v2/timer",
        method="POST",
        json={
            "timer": V2_TRANSFER_TIMER,
        },
        status=201,
        metadata={
            "timer_id": TIMER_ID,
            "source_endpoint": SOURCE_EP_ID,
            "destination_endpoint": DEST_EP_ID,
        },
    ),
    flow_timer_success=RegisteredResponse(
        service="timer",
        path="/v2/timer",
        method="POST",
        json={
            "timer": V2_FLOW_TIMER,
        },
        status=201,
        metadata={
            "timer_id": V2_FLOW_TIMER["job_id"],
            "flow_id": FLOW_ID,
            "callback_body": V2_FLOW_TIMER["callback_body"],
            "schedule": V2_FLOW_TIMER["schedule"],
        },
    ),
)



import sys

import pytest
import responses

import globus_sdk
import globus_sdk.token_storage
from globus_sdk.testing import RegisteredResponse, load_response

# the JWT will have a client ID in its audience claim
# make sure to use that value when trying to decode it
CLIENT_ID_FROM_JWT = "7fb58e00-839d-44e3-8047-10a502612dca"


class InfiniteLeewayDecoder(globus_sdk.IDTokenDecoder):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.jwt_leeway = sys.maxsize


@pytest.fixture(autouse=True)
def _mock_input(capsys, monkeypatch):
    def _fake_input(s):
        print(s)
        return "mock_input"

    monkeypatch.setattr("builtins.input", _fake_input)


@pytest.fixture
def oidc_and_jwk_responses():
    RegisteredResponse(
        service="auth",
        path="/.well-known/openid-configuration",
        method="GET",
        json={
            "issuer": "https://auth.globus.org",
            "authorization_endpoint": "https://auth.globus.org/v2/oauth2/authorize",
            "userinfo_endpoint": "https://auth.globus.org/v2/oauth2/userinfo",
            "token_endpoint": "https://auth.globus.org/v2/oauth2/token",
            "revocation_endpoint": "https://auth.globus.org/v2/oauth2/token/revoke",
            "jwks_uri": "https://auth.globus.org/jwk.json",
            "response_types_supported": [
                "code",
                "token",
                "token id_token",
                "id_token",
            ],
            "id_token_signing_alg_values_supported": ["RS512"],
            "scopes_supported": ["openid", "email", "profile"],
            "token_endpoint_auth_methods_supported": ["client_secret_basic"],
            "claims_supported": [
                "at_hash",
                "aud",
                "email",
                "exp",
                "name",
                "nonce",
                "preferred_username",
                "iat",
                "iss",
                "sub",
            ],
            "subject_types_supported": ["public"],
        },
    ).add()
    RegisteredResponse(
        service="auth",
        path="/jwk.json",
        method="GET",
        json={
            "keys": [
                {
                    "alg": "RS512",
                    "e": "AQAB",
                    "kty": "RSA",
                    "n": "73l27Yp7WT2c0Ve7EoGJ13AuKzg-GHU7Mpgx0JKa_hO04gAXSVXRadQy7gmdLLtAK8uBVcV0fHGgsBl4J92t-I7hayiJSLbgbX-sZhI_OfegeOLcSNB9poPS9w60XGqR9buYOW2x-KXXitsmyHXNmg_-1u0uqfKHu9pmST8dcjUYXTM5F3oJpQKeJlSH8daMlDks4xb9Y83EEFRv-ppY965-WTm2NW4pwLlbgGTWFvZ6YS6GTb-mfGwGuzStI0lKZ7dOFx9ryYQ4wSoUVHtIrypT-gbuaT90Z2SkwOH-GaEZJkudctBeGpieOsyC7P40UXpwgGNFy3xoWL4vHpnHmQ",  # noqa: E501
                    "use": "sig",
                }
            ]
        },
    ).add()


@pytest.fixture
def token_response():
    RegisteredResponse(
        service="auth",
        path="/v2/oauth2/token",
        method="POST",
        status=200,
        json={
            "access_token": "auth_access_token",
            "scope": "openid",
            "expires_in": 172800,
            "token_type": "Bearer",
            "resource_server": "auth.globus.org",
            # this is a real ID token
            # and since the JWK is real as well, it should decode correctly
            # but it will have a bad expiration time because it's a very old token
            "id_token": "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzUxMiJ9.eyJzdWIiOiJjOGFhZDQzZS1kMjc0LTExZTUtYmY5OC04YjAyODk2Y2Y3ODIiLCJvcmdhbml6YXRpb24iOiJHbG9idXMiLCJuYW1lIjoiU3RlcGhlbiBSb3NlbiIsInByZWZlcnJlZF91c2VybmFtZSI6InNpcm9zZW4yQGdsb2J1c2lkLm9yZyIsImlkZW50aXR5X3Byb3ZpZGVyIjoiNDExNDM3NDMtZjNjOC00ZDYwLWJiZGItZWVlY2FiYTg1YmQ5IiwiaWRlbnRpdHlfcHJvdmlkZXJfZGlzcGxheV9uYW1lIjoiR2xvYnVzIElEIiwiZW1haWwiOiJzaXJvc2VuQHVjaGljYWdvLmVkdSIsImxhc3RfYXV0aGVudGljYXRpb24iOjE2MjE0ODEwMDYsImlkZW50aXR5X3NldCI6W3sic3ViIjoiYzhhYWQ0M2UtZDI3NC0xMWU1LWJmOTgtOGIwMjg5NmNmNzgyIiwib3JnYW5pemF0aW9uIjoiR2xvYnVzIiwibmFtZSI6IlN0ZXBoZW4gUm9zZW4iLCJ1c2VybmFtZSI6InNpcm9zZW4yQGdsb2J1c2lkLm9yZyIsImlkZW50aXR5X3Byb3ZpZGVyIjoiNDExNDM3NDMtZjNjOC00ZDYwLWJiZGItZWVlY2FiYTg1YmQ5IiwiaWRlbnRpdHlfcHJvdmlkZXJfZGlzcGxheV9uYW1lIjoiR2xvYnVzIElEIiwiZW1haWwiOiJzaXJvc2VuQHVjaGljYWdvLmVkdSIsImxhc3RfYXV0aGVudGljYXRpb24iOjE2MjE0ODEwMDZ9LHsic3ViIjoiYjZlMjI3ZTgtZGI1Mi0xMWU1LWI2ZmYtYzNiMWNjMjU5ZTBkIiwibmFtZSI6bnVsbCwidXNlcm5hbWUiOiJzaXJvc2VuK2JhZGVtYWlsQGdsb2J1cy5vcmciLCJpZGVudGl0eV9wcm92aWRlciI6IjkyN2Q3MjM4LWY5MTctNGViMi05YWNlLWM1MjNmYTliYTM0ZSIsImlkZW50aXR5X3Byb3ZpZGVyX2Rpc3BsYXlfbmFtZSI6Ikdsb2J1cyBTdGFmZiIsImVtYWlsIjoic2lyb3NlbitiYWRlbWFpbEBnbG9idXMub3JnIiwibGFzdF9hdXRoZW50aWNhdGlvbiI6bnVsbH0seyJzdWIiOiJmN2Y4OWQwYS1kYzllLTExZTUtYWRkMC1hM2NiZDFhNTU5YjMiLCJuYW1lIjpudWxsLCJ1c2VybmFtZSI6InNpcm9zZW4rYmFkZW1haWwyQGdsb2J1cy5vcmciLCJpZGVudGl0eV9wcm92aWRlciI6IjkyN2Q3MjM4LWY5MTctNGViMi05YWNlLWM1MjNmYTliYTM0ZSIsImlkZW50aXR5X3Byb3ZpZGVyX2Rpc3BsYXlfbmFtZSI6Ikdsb2J1cyBTdGFmZiIsImVtYWlsIjoic2lyb3NlbitiYWRlbWFpbDJAZ2xvYnVzLm9yZyIsImxhc3RfYXV0aGVudGljYXRpb24iOm51bGx9XSwiaXNzIjoiaHR0cHM6Ly9hdXRoLmdsb2J1cy5vcmciLCJhdWQiOiI3ZmI1OGUwMC04MzlkLTQ0ZTMtODA0Ny0xMGE1MDI2MTJkY2EiLCJleHAiOjE2MjE2NTM4MTEsImlhdCI6MTYyMTQ4MTAxMSwiYXRfaGFzaCI6IjFQdlVhbmNFdUxfc2cxV1BsNWx1TUVGR2tjTDZQaDh1cWdpVUZzejhkZUEifQ.CtfnFtfM32ICo0euHv9GnpVHFL1jWz0NriPTXAv6w08Ylk9JBJtmB3oMKNSO-1TGoWUPFDp9TFFk6N32VyF0hsVDtT5DT3t5oq0qfqbPrZA3R04HARW0xtcK_ejNDHBmj6wysey3EzjT764XTvcGOe63CKQ_RJm97ulVaseIT0Aet7AYo5tQuOiSOQ70xzL7Oax3W6TrWi3FIAA-PIMSrAJKbsG7imGOVkaIObG9a-X5yTOcrB4IG4Wat-pN_QiCiiOw_LDCF-r455PwalmnSGUugMYfsdL2k3UxqwOMLIppHnx5-UVAzj3mygj8eZTp6imjqxNMdakS3vhG8dtxbw",  # noqa: E501
            "other_tokens": [],
        },
    ).add()


def test_globus_app_only_gets_oidc_data_once(
    oidc_and_jwk_responses, token_response, monkeypatch
):
    # needed for logout later
    load_response(globus_sdk.NativeAppAuthClient.oauth2_revoke_token)

    def _count_oidc_calls():
        calls = [c.request for c in responses.calls]
        calls = [
            c
            for c in calls
            if c.url == "https://auth.globus.org/.well-known/openid-configuration"
        ]
        return len(calls)

    def _count_jwk_calls():
        calls = [c.request for c in responses.calls]
        calls = [c for c in calls if c.url == "https://auth.globus.org/jwk.json"]
        return len(calls)

    memory_storage = globus_sdk.token_storage.MemoryTokenStorage()
    config = globus_sdk.GlobusAppConfig(
        token_storage=memory_storage,
        id_token_decoder=InfiniteLeewayDecoder,
    )
    user_app = globus_sdk.UserApp(
        "test-app", config=config, client_id=CLIENT_ID_FROM_JWT
    )

    # start: we haven't made any calls yet
    assert _count_oidc_calls() == 0
    assert _count_jwk_calls() == 0

    # after login: we made one call for each
    user_app.login()
    assert _count_oidc_calls() == 1
    assert _count_jwk_calls() == 1

    # logout and confirm that it was effective
    user_app.logout()
    assert user_app.login_required() is True

    # second login: call counts did not increase
    user_app.login()
    assert _count_oidc_calls() == 1
    assert _count_jwk_calls() == 1


def test_globus_app_can_set_custom_id_token_decoder_via_config_provider(
    oidc_and_jwk_responses,
    token_response,
):
    init_counter = 0

    class CustomDecoder(InfiniteLeewayDecoder):
        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args, **kwargs)
            nonlocal init_counter
            init_counter += 1

    memory_storage = globus_sdk.token_storage.MemoryTokenStorage()
    config = globus_sdk.GlobusAppConfig(
        token_storage=memory_storage,
        id_token_decoder=CustomDecoder,
    )
    user_app = globus_sdk.UserApp(
        "test-app", config=config, client_id=CLIENT_ID_FROM_JWT
    )

    # confirm that our custom init got called during app init
    assert init_counter == 1
    # login, and confirm no failure (the default would fail on the stale id_token used)
    user_app.login()


def test_globus_app_can_set_custom_id_token_decoder_via_config_instance(
    oidc_and_jwk_responses,
    token_response,
):
    # needed for logout later
    load_response(globus_sdk.NativeAppAuthClient.oauth2_revoke_token)

    call_counter = 0

    class CustomDecoder(InfiniteLeewayDecoder):
        def decode(self, *args, **kwargs) -> None:
            nonlocal call_counter
            call_counter += 1
            return super().decode(*args, **kwargs)

    login_client = globus_sdk.NativeAppAuthClient(client_id=CLIENT_ID_FROM_JWT)
    memory_storage = globus_sdk.token_storage.MemoryTokenStorage()
    config = globus_sdk.GlobusAppConfig(
        token_storage=memory_storage, id_token_decoder=CustomDecoder(login_client)
    )
    user_app = globus_sdk.UserApp("test-app", config=config, login_client=login_client)

    assert call_counter == 0
    # login, and confirm no failure (the default would fail on the stale id_token used)
    user_app.login()

    # confirm that our custom decode got called
    assert call_counter == 1

    # logout and log back in; did it get called again? good
    user_app.logout()
    user_app.login()
    assert call_counter == 2


def test_globus_app_custom_id_token_decoder_instance_can_overload_jwt_leeway(
    oidc_and_jwk_responses,
    token_response,
):
    # needed for logout later
    load_response(globus_sdk.NativeAppAuthClient.oauth2_revoke_token)

    login_client = globus_sdk.NativeAppAuthClient(client_id=CLIENT_ID_FROM_JWT)
    memory_storage = globus_sdk.token_storage.MemoryTokenStorage()
    config = globus_sdk.GlobusAppConfig(
        token_storage=memory_storage,
        id_token_decoder=globus_sdk.IDTokenDecoder(
            login_client, jwt_leeway=sys.maxsize
        ),
    )
    user_app = globus_sdk.UserApp("test-app", config=config, login_client=login_client)

    # login, and confirm no failure (the default would fail on the stale id_token used)
    user_app.login()



import json

import pytest

from globus_sdk import MISSING, FlowsAPIError
from globus_sdk.testing import get_last_request, load_response


@pytest.mark.parametrize("input_schema", [MISSING, {}])
def test_validate_flow(flows_client, input_schema):
    metadata = load_response(flows_client.validate_flow).metadata

    # Prepare the payload
    payload = {"definition": metadata["success"]}
    if input_schema is not MISSING:
        payload["input_schema"] = input_schema

    resp = flows_client.validate_flow(**payload)
    assert resp.data["scopes"] == {
        "User": ["urn:globus:auth:scope:transfer.api.globus.org:all"]
    }

    # Check what was actually sent
    last_req = get_last_request()
    req_body = json.loads(last_req.body)
    # Ensure the input schema is not sent if omitted
    assert req_body == payload


def test_validate_flow_error_parsing(flows_client):
    metadata = load_response(
        flows_client.validate_flow, case="definition_error"
    ).metadata

    # Make sure we get an error response
    with pytest.raises(FlowsAPIError) as excinfo:
        flows_client.validate_flow(definition=metadata["invalid"])

    err = excinfo.value
    assert err.code == "UNPROCESSABLE_ENTITY"
    assert err.messages == [
        (
            "A state of type 'Action' must be defined as either terminal "
            '("End": true) or transitional ("Next": "NextStateId")'
        ),
    ]



import pytest

from globus_sdk import GlobusSDKUsageError
from globus_sdk.token_storage.base import _slugify_app_name


@pytest.mark.parametrize(
    "app_name, expected_slug",
    (
        ("Globus CLI v3.30.0", "globus-cli-v3-30-0"),
        ("my-cool-app", "my-cool-app"),
        ("jonathan", "jonathan"),
    ),
)
def test_app_name_slugification_of_basic_app_names(app_name, expected_slug):
    actual_slug = _slugify_app_name(app_name)
    assert actual_slug == expected_slug


def test_app_name_slugification_replaces_reserved_characters():
    app_name = 'glob<>:"/\\|?*us'
    expected = "glob+++++++++us"
    actual_slug = _slugify_app_name(app_name)
    assert actual_slug == expected


def test_app_name_slugification_removes_control_characters():
    app_name = "glob\0\n\t\rus"
    expected = "globus"
    actual_slug = _slugify_app_name(app_name)
    assert actual_slug == expected


def test_app_name_slugification_rejects_empty_name():
    with pytest.raises(GlobusSDKUsageError, match="name results in the empty string"):
        _slugify_app_name("\n\t\r")


@pytest.mark.parametrize("app_name", ("CON", "PRN\n", "nul", "LPT9", "COM1"))
def test_app_name_slugification_rejects_reserved_names(app_name):
    # # https://stackoverflow.com/a/31976060
    with pytest.raises(GlobusSDKUsageError, match="reserved filename"):
        _slugify_app_name(app_name)


def test_app_name_allows_reserved_name_prefixes():
    # Prefix of app name ("CON") is reserved, but the slugified name is not.
    app_name = "Continuous Deployment App"
    expected = "continuous-deployment-app"
    actual_slug = _slugify_app_name(app_name)
    assert actual_slug == expected



from __future__ import annotations

import typing as t
import uuid

from globus_sdk import response

from .client import GroupsClient
from .data import (
    _GROUP_MEMBER_VISIBILITY_T,
    _GROUP_REQUIRED_SIGNUP_FIELDS_T,
    _GROUP_ROLE_T,
    _GROUP_VISIBILITY_T,
    BatchMembershipActions,
    GroupPolicies,
)


class GroupsManager:
    """
    A wrapper for the groups client with common membership and group actions
    wrapped in convenient methods with parameters and type hints.

    .. automethodlist:: globus_sdk.GroupsManager
    """

    def __init__(self, client: GroupsClient | None = None) -> None:
        self.client = client or GroupsClient()

    def create_group(
        self,
        name: str,
        description: str,
        *,
        parent_id: uuid.UUID | str | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Create a group with the given name.  If a parent ID is included, the
        group will be a subgroup of the given parent group.

        :param name: The name of the group
        :param description: A description of the group
        :param parent_id: The ID of the parent group, if there is one
        """
        data = {
            "name": name,
            "description": description,
            "parent_id": str(parent_id) if parent_id is not None else None,
        }
        return self.client.create_group(data=data)

    def set_group_policies(
        self,
        group_id: uuid.UUID | str,
        *,
        is_high_assurance: bool,
        group_visibility: _GROUP_VISIBILITY_T,
        group_members_visibility: _GROUP_MEMBER_VISIBILITY_T,
        join_requests: bool,
        signup_fields: t.Iterable[_GROUP_REQUIRED_SIGNUP_FIELDS_T],
        authentication_assurance_timeout: int | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Set the group policies for the given group.

        :param group_id: The ID of the group on which to set policies
        :param is_high_assurance: Whether the group can provide a High Assurance
            guarantee when used for access controls
        :param group_visibility: The visibility of the group
        :param group_members_visibility: The visibility of memberships within the group
        :param join_requests: Whether the group allows users to request to join
        :param signup_fields: The required fields for a user to sign up for the group
        :param authentication_assurance_timeout: The timeout used when this group is
            used to apply a High Assurance authentication guarantee
        """
        data = GroupPolicies(
            is_high_assurance=is_high_assurance,
            group_visibility=group_visibility,
            group_members_visibility=group_members_visibility,
            join_requests=join_requests,
            signup_fields=signup_fields,
            authentication_assurance_timeout=authentication_assurance_timeout,
        )
        return self.client.set_group_policies(group_id, data=data)

    def accept_invite(
        self, group_id: uuid.UUID | str, identity_id: uuid.UUID | str
    ) -> response.GlobusHTTPResponse:
        """
        Accept invite for an identity.  The identity must belong to
        the identity set of the authenticated user.

        :param group_id: The ID of the group
        :param identity_id: The identity for whom to accept the invite
        """
        actions = BatchMembershipActions().accept_invites([identity_id])
        return self.client.batch_membership_action(group_id, actions)

    def add_member(
        self,
        group_id: uuid.UUID | str,
        identity_id: uuid.UUID | str,
        *,
        role: _GROUP_ROLE_T = "member",
    ) -> response.GlobusHTTPResponse:
        """
        Add an identity to a group with the given role.

        :param group_id: The ID of the group
        :param identity_id: The identity to add to the group
        :param role: The role for the new group member
        """
        actions = BatchMembershipActions().add_members([identity_id], role=role)
        return self.client.batch_membership_action(group_id, actions)

    def approve_pending(
        self, group_id: uuid.UUID | str, identity_id: uuid.UUID | str
    ) -> response.GlobusHTTPResponse:
        """
        Approve an identity with a pending join request.

        :param group_id: The ID of the group
        :param identity_id: The identity to approve as a member of the group
        """
        actions = BatchMembershipActions().approve_pending([identity_id])
        return self.client.batch_membership_action(group_id, actions)

    def change_role(
        self,
        group_id: uuid.UUID | str,
        identity_id: uuid.UUID | str,
        role: _GROUP_ROLE_T,
    ) -> response.GlobusHTTPResponse:
        """
        Change the role of the given identity in the given group.

        :param group_id: The ID of the group
        :param identity_id: The identity to assign the *role* to
        :param role: The role that will be assigned to the *identity_id*
        """
        actions = BatchMembershipActions().change_roles(role, [identity_id])
        return self.client.batch_membership_action(group_id, actions)

    def decline_invite(
        self, group_id: uuid.UUID | str, identity_id: uuid.UUID | str
    ) -> response.GlobusHTTPResponse:
        """
        Decline an invitation for a given identity.

        :param group_id: The ID of the group
        :param identity_id: The identity for whom to decline the invitation
        """
        actions = BatchMembershipActions().decline_invites([identity_id])
        return self.client.batch_membership_action(group_id, actions)

    def invite_member(
        self,
        group_id: uuid.UUID | str,
        identity_id: uuid.UUID | str,
        *,
        role: _GROUP_ROLE_T = "member",
    ) -> response.GlobusHTTPResponse:
        """
        Invite an identity to a group with the given role.

        :param group_id: The ID of the group
        :param identity_id: The identity to invite as a new group member
        :param role: The role for the invited group member
        """
        actions = BatchMembershipActions().invite_members([identity_id], role=role)
        return self.client.batch_membership_action(group_id, actions)

    def join(
        self, group_id: uuid.UUID | str, identity_id: uuid.UUID | str
    ) -> response.GlobusHTTPResponse:
        """
        Join a group with the given identity.  The identity must be in the
        authenticated users identity set.

        :param group_id: The ID of the group
        :param identity_id: The identity to use to join the group
        """
        actions = BatchMembershipActions().join([identity_id])
        return self.client.batch_membership_action(group_id, actions)

    def leave(
        self, group_id: uuid.UUID | str, identity_id: uuid.UUID | str
    ) -> response.GlobusHTTPResponse:
        """
        Leave a group that one of the identities in the authenticated user's
        identity set is a member of.

        :param group_id: The ID of the group
        :param identity_id: The identity to remove from the group
        """
        actions = BatchMembershipActions().leave([identity_id])
        return self.client.batch_membership_action(group_id, actions)

    def reject_join_request(
        self, group_id: uuid.UUID | str, identity_id: uuid.UUID | str
    ) -> response.GlobusHTTPResponse:
        """
        Reject a member that has requested to join the group.

        :param group_id: The ID of the group
        :param identity_id: The identity to reject from the group
        """
        actions = BatchMembershipActions().reject_join_requests([identity_id])
        return self.client.batch_membership_action(group_id, actions)

    def remove_member(
        self, group_id: uuid.UUID | str, identity_id: uuid.UUID | str
    ) -> response.GlobusHTTPResponse:
        """
        Remove a member from a group.  This must be done as an admin or manager
        of the group.

        :param group_id: The ID of the group
        :param identity_id: The identity to remove from the group
        """
        actions = BatchMembershipActions().remove_members([identity_id])
        return self.client.batch_membership_action(group_id, actions)

    def request_join(
        self, group_id: uuid.UUID | str, identity_id: uuid.UUID | str
    ) -> response.GlobusHTTPResponse:
        """
        Request to join a group.

        :param group_id: The ID of the group
        :param identity_id: The identity to use to request membership in the group
        """
        actions = BatchMembershipActions().request_join([identity_id])
        return self.client.batch_membership_action(group_id, actions)



import re

import pytest


def test_listknownscopes_rejects_wrong_object_type(sphinx_runner, capsys):
    sphinx_runner.ensure_failure(
        ".. listknownscopes:: globus_sdk.BaseClient",
    )

    captured = capsys.readouterr()
    err_lines = captured.err.splitlines()

    test_line = None
    for line in err_lines:
        if "scope collection" in line:
            test_line = line
            break
    else:
        pytest.fail("Didn't find 'scope collection' in stderr")

    assert re.search(
        r"Expected <class '(\w+\.)+BaseClient'> to be a scope collection", test_line
    )


# choose an arbitrary scope collection from the SDK and confirm that listknownscopes
# will render its list of members
# for this case, we're using `TimersScopes`
def test_listknownscopes_of_timers(sphinx_runner):
    etree = sphinx_runner.to_etree(
        ".. listknownscopes:: globus_sdk.scopes.TimersScopes",
    )

    assert etree.tag == "document"

    paragraphs = etree.findall("paragraph")
    assert len(paragraphs) == 2
    paragraph0, paragraph1 = paragraphs

    assert paragraph0.text.startswith(
        "Various scopes are available as attributes of this object."
    )

    console_block = etree.find("doctest_block")
    assert console_block is not None
    assert console_block.text == ">>> TimersScopes.timer"

    emphasized_text = paragraph1.find("strong")
    assert emphasized_text is not None
    assert emphasized_text.text == "Supported Scopes"

    scope_list = etree.find("bullet_list")
    assert scope_list is not None
    scope_items = scope_list.findall("./list_item/paragraph/literal")
    assert len(scope_items) == 1
    assert scope_items[0].text == "timer"


def test_listknownscopes_of_timers_with_forced_example(sphinx_runner):
    etree = sphinx_runner.to_etree(
        """\
        .. listknownscopes:: globus_sdk.scopes.TimersScopes
            :example_scope: frobulate
        """,
    )

    assert etree.tag == "document"

    console_block = etree.find("doctest_block")
    assert console_block is not None
    assert console_block.text == ">>> TimersScopes.frobulate"


def test_listknownscopes_of_timers_with_altered_basename(sphinx_runner):
    etree = sphinx_runner.to_etree(
        """\
        .. listknownscopes:: globus_sdk.scopes.TimersScopes
            :base_name: ScopeMuddler
        """,
    )

    assert etree.tag == "document"

    console_block = etree.find("doctest_block")
    assert console_block is not None
    assert console_block.text == ">>> ScopeMuddler.timer"



import pathlib
import shutil
import tempfile
import textwrap
from xml.etree import ElementTree

import pytest
import responses

HERE = pathlib.Path(__file__).parent
REPO_ROOT = HERE.parent.parent.parent


#
# sphinx's intersphinx plugin loads an inventory from the docs.python.org docs for
# linkage each time it runs
# we use a cached copy so that:
# - tests don't require the network
# - we don't put unnecessary load on the python docs site
# - tests will be faster
#
@pytest.fixture(autouse=True)
def _mock_intersphinx_object_inventory_response(mocked_responses):
    objects_inv = HERE / "objects.inv"
    responses.get(
        url="https://docs.python.org/3/objects.inv", body=objects_inv.read_bytes()
    )


class SingleFileSphinxRunner:
    """
    A SingleFileSphinxRunner runs sphinx with the full globus-sdk config loaded, against
    some target RST source, and returns the parsed XML it produces.

    For content with sphinx-specific features like `py:meth` and other sphinx domain
    features, this is the best way of ensuring we get a realistic build with good
    outputs.

    Primarily, usage should use ``to_etree`` like so:

    >>> def test_my_thing(sphinx_runner):
    >>>     etree_element = sphinx_runner.to_etree(
    >>>         ":py:class:`globus_sdk.TimersClient`"
    >>>     )

    Because it's often difficult to know the structure of the produced XML, if
    `debug=True` is passed, this will print the full XML to stdout, for you to
    see in your test outputs.
    """

    def to_etree(self, content, dedent=True, debug=False):
        from sphinx.cmd.build import build_main as sphinx_main

        if dedent:
            content = textwrap.dedent(content)

        source_dir = tempfile.TemporaryDirectory()
        out_dir = tempfile.TemporaryDirectory()
        with source_dir, out_dir:
            self._prepare_file(source_dir.name, content, "index.rst")
            self._prepare_sphinx_config(source_dir.name)

            sphinx_rc = sphinx_main([source_dir.name, out_dir.name, "-b", "xml"])

            assert sphinx_rc == 0

            output_xml = pathlib.Path(out_dir.name) / "index.xml"
            xml_text = output_xml.read_text(encoding="utf-8")
            if debug:
                print("--- debug from sphinx runner ---")
                print()
                print(xml_text)
                print()
                print("--- end debug from sphinx runner ---")
            return ElementTree.fromstring(xml_text)

    def ensure_failure(self, content, dedent=True, debug=False):
        from sphinx.cmd.build import build_main as sphinx_main

        if dedent:
            content = textwrap.dedent(content)

        source_dir = tempfile.TemporaryDirectory()
        out_dir = tempfile.TemporaryDirectory()
        with source_dir, out_dir:
            self._prepare_file(source_dir.name, content, "index.rst")
            self._prepare_sphinx_config(source_dir.name)

            sphinx_rc = sphinx_main([source_dir.name, out_dir.name, "-b", "xml"])
            assert sphinx_rc != 0

    def _prepare_file(self, source_dir, content, filename):
        source_path = pathlib.Path(source_dir) / filename
        source_path.write_text(content, encoding="utf-8")

    def _prepare_sphinx_config(self, source_dir):
        shutil.copy(
            REPO_ROOT / "docs" / "conf.py", pathlib.Path(source_dir) / "conf.py"
        )


class DocutilsRunner:
    """
    A DocutilsRunner is a direct user of docutils, with no special sphinx customizations
    applied.

    It can fail on sphinx-specific content, but it lets us directly test underlying
    docutils behaviors.

    Primarily, usage should use ``to_etree`` like so:

    >>> def test_my_thing(docutils_runner):
    >>>     etree_element = docutils_runner.to_etree(
    >>>         ".. note:: a note directive"
    >>>     )

    Because it's often difficult to know the structure of the produced XML, if
    `debug=True` is passed, this will print the full XML to stdout, for you to
    see in your test outputs.
    """

    def __init__(self) -> None:
        from docutils.frontend import get_default_settings
        from docutils.parsers.rst import Parser

        self.parser = Parser()
        self.settings = get_default_settings(self.parser)
        # set the halt_level so that docutils won't "power through" noncritical errors
        #   https://docutils.sourceforge.io/docs/user/config.html#halt-level
        self.settings.halt_level = 2

    def to_etree(self, content, dedent=True, debug=False):

        if dedent:
            content = textwrap.dedent(content)

        doc = self.new_doc()
        xml_string = self.make_xml(content, doc)
        if debug:
            print("--- debug from docutils runner ---")
            print()
            print(xml_string)
            print()
            print("--- end debug from docutils runner ---")
        etree = ElementTree.fromstring(xml_string)
        return etree

    def new_doc(self, doc_source="TEST"):
        from docutils.utils import new_document

        return new_document(doc_source, self.settings.copy())

    def make_xml(self, source, doc):
        # docutils produces 'xml.dom.minidom'
        # for a nicer API, convert to XML so we can reparse as etree
        self.parser.parse(source, doc)
        dom = doc.asdom()
        xml_string = dom.toxml()
        # also, unlink the dom object to ensure we don't get cyclic references
        # which make GC fail to collect these (see xml.dom.minidom docs for detail)
        dom.unlink()
        return xml_string


@pytest.fixture
def sphinx_runner():
    pytest.importorskip("sphinx", reason="testing sphinx extension needs sphinx")

    return SingleFileSphinxRunner()


@pytest.fixture
def docutils_runner():
    pytest.importorskip("docutils", reason="testing sphinx extension needs docutils")
    return DocutilsRunner()


@pytest.fixture
def sphinxext():
    """
    Provide the extension subpackage as a fixture so that we can properly capture
    skip requirements and avoid awkward import requirements.
    """
    pytest.importorskip("docutils", reason="testing sphinx extension needs docutils")

    import globus_sdk._internal.extensions.sphinxext

    return globus_sdk._internal.extensions.sphinxext


@pytest.fixture
def register_temporary_directive():
    pytest.importorskip("docutils", reason="testing sphinx extension needs docutils")
    from docutils.parsers.rst.directives import _directives, register_directive

    registered_names = []

    def func(name, directive):
        registered_names.append(name)
        register_directive(name, directive)

    yield func

    for name in registered_names:
        del _directives[name]



import json

import pytest

import globus_sdk
from globus_sdk.testing import get_last_request, load_response


@pytest.fixture
def search_client():
    client = globus_sdk.SearchClient()
    with client.retry_config.tune(max_retries=0):
        yield client


def test_search_role_create(search_client):
    meta = load_response(search_client.create_role).metadata
    send_data = {
        "role_name": meta["role_name"],
        "principal": "urn:globus:auth:identity:" + meta["identity_id"],
    }

    res = search_client.create_role(meta["index_id"], send_data)
    assert res.http_status == 200
    assert res["index_id"] == meta["index_id"]
    assert res["role_name"] == "writer"

    last_req = get_last_request()
    sent = json.loads(last_req.body)
    assert sent == send_data


def test_search_role_delete(search_client):
    meta = load_response(search_client.delete_role).metadata

    res = search_client.delete_role(meta["index_id"], meta["role_id"])
    assert res.http_status == 200
    assert res["success"] is True
    assert res["deleted"]["index_id"] == meta["index_id"]
    assert res["deleted"]["id"] == meta["role_id"]


def test_search_role_list(search_client):
    meta = load_response(search_client.get_role_list).metadata

    res = search_client.get_role_list(meta["index_id"])
    assert res.http_status == 200
    role_list = res["role_list"]
    assert isinstance(role_list, list)
    assert len(role_list) == 2



import typing as t
import uuid

from responses.matchers import json_params_matcher

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

POLICY_REQUEST_ARGS = {
    "project_id": str(uuid.uuid1()),
    "display_name": "Policy of Foo",
    "description": "Controls access to Foo",
}


def make_request_body(request_args: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    request_body = request_args.copy()
    request_body["project_id"] = str(request_args["project_id"])

    for domain_constraints in [
        "domain_constraints_include",
        "domain_constraints_exclude",
    ]:
        if domain_constraints in request_args:
            request_body[domain_constraints] = request_args[domain_constraints]

    return request_body


def make_response_body(request_args: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    response_body = request_args.copy()
    response_body["project_id"] = str(request_args["project_id"])
    response_body["id"] = str(uuid.uuid1())

    for domain_constraints in [
        "domain_constraints_include",
        "domain_constraints_exclude",
    ]:
        if domain_constraints in request_args:
            response_body[domain_constraints] = request_args.get(domain_constraints)

    return response_body


def register_response(
    args: t.Mapping[str, t.Any],
    match: t.Any = None,
) -> RegisteredResponse:
    request_args = {**POLICY_REQUEST_ARGS, **args}
    request_body = make_request_body(request_args)
    response_body = make_response_body(request_args)

    return RegisteredResponse(
        service="auth",
        method="POST",
        path="/v2/api/policies",
        json={"policy": response_body},
        metadata={
            # Test functions use 'args' to form request
            "args": request_args,
            # Test functions use 'response' to verify response
            "response": response_body,
        },
        match=(
            [json_params_matcher({"policy": request_body})] if match is None else match
        ),
    )


RESPONSES = ResponseSet(
    default=register_response({}, match=[]),
    project_id_str=register_response({"project_id": str(uuid.uuid1())}),
    project_id_uuid=register_response({"project_id": uuid.uuid1()}),
    high_assurance=register_response(
        {
            "high_assurance": True,
            "authentication_assurance_timeout": 35,
            "required_mfa": False,
        }
    ),
    not_high_assurance=register_response(
        {"high_assurance": False, "required_mfa": False}
    ),
    required_mfa=register_response(
        {
            "high_assurance": True,
            "authentication_assurance_timeout": 45,
            "required_mfa": True,
        }
    ),
    authentication_assurance_timeout=register_response(
        {"authentication_assurance_timeout": 23}
    ),
    display_name=register_response(
        {"display_name": str(uuid.uuid4()).replace("-", "")}
    ),
    description=register_response({"description": str(uuid.uuid4()).replace("-", "")}),
    domain_constraints_include=register_response(
        {
            "domain_constraints_include": ["globus.org", "uchicago.edu"],
        },
    ),
    empty_domain_constraints_include=register_response(
        {
            "domain_constraints_include": [],
        },
    ),
    no_domain_constraints_include=register_response(
        {
            "domain_constraints_include": None,
        },
    ),
    domain_constraints_exclude=register_response(
        {
            "domain_constraints_exclude": ["globus.org", "uchicago.edu"],
        },
    ),
    empty_domain_constraints_exclude=register_response(
        {
            "domain_constraints_exclude": [],
        },
    ),
    no_domain_constraints_exclude=register_response(
        {
            "domain_constraints_exclude": None,
        },
    ),
)



import pytest

from globus_sdk import GlobusSDKUsageError
from globus_sdk.testing import load_response


def test_get_scopes(service_client):
    meta = load_response(service_client.get_scopes).metadata
    res = service_client.get_scopes()

    assert {scope["id"] for scope in res["scopes"]} == set(meta["scope_ids"])


def test_get_scopes_by_ids(service_client):
    meta = load_response(service_client.get_scopes, case="id").metadata
    res = service_client.get_scopes(ids=[meta["scope_id"]])

    assert res["scopes"][0]["id"] == meta["scope_id"]


def test_get_scopes_by_strings(service_client):
    meta = load_response(service_client.get_scopes, case="string").metadata
    res = service_client.get_scopes(scope_strings=[meta["scope_string"]])

    assert res["scopes"][0]["scope_string"] == meta["scope_string"]


def test_get_scopes_id_strings_mutually_exclusive(service_client):
    with pytest.raises(GlobusSDKUsageError):
        service_client.get_scopes(
            scope_strings=["foo"],
            ids=["18a8cd00-700a-4fcb-b6da-6efca558c369"],
        )



"""
Inspect the signatures of paginated methods and compare them against their
attached paginator requirements.
"""

import inspect

import pytest

import globus_sdk

_CLIENTS_TO_CHECK = []
for attrname in dir(globus_sdk):
    obj = getattr(globus_sdk, attrname)
    if obj is globus_sdk.BaseClient:
        continue
    if isinstance(obj, type) and issubclass(obj, globus_sdk.BaseClient):
        _CLIENTS_TO_CHECK.append(obj)

_METHODS_TO_CHECK = []
for cls in _CLIENTS_TO_CHECK:
    methods = inspect.getmembers(cls, predicate=inspect.isfunction)
    for name, value in methods:
        if name.startswith("_"):
            continue
        # inherited, non-overloaded methods
        if name not in cls.__dict__:
            continue
        if getattr(value, "_has_paginator", False):
            _METHODS_TO_CHECK.append(value)


@pytest.mark.parametrize("method", _METHODS_TO_CHECK)
def test_paginated_method_matches_paginator_requirements(method):
    paginator_class = method._paginator_class

    sig = inspect.signature(method)
    kwarg_names = {
        p.name
        for p in sig.parameters.values()
        if p.kind in (p.POSITIONAL_OR_KEYWORD, p.KEYWORD_ONLY)
    }

    for param_name in paginator_class._REQUIRES_METHOD_KWARGS:
        assert param_name in kwarg_names, method.__qualname__



import globus_sdk


def uses_data_access(
    transfer_client: globus_sdk.TransferClient, collection_id: str
) -> bool:
    """
    Use the provided `transfer_client` to lookup a collection by ID.

    Based on the record, return `True` if it uses a `data_access` scope and `False`
    otherwise.
    """
    doc = transfer_client.get_endpoint(collection_id)
    if doc["entity_type"] != "GCSv5_mapped_collection":
        return False
    if doc["high_assurance"]:
        return False
    return True


# Tutorial Client ID - <replace this with your own client>
NATIVE_CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"

# Globus Tutorial Collection 1 & 2
# https://app.globus.org/file-manager/collections/6c54cade-bde5-45c1-bdea-f4bd71dba2cc
# https://app.globus.org/file-manager/collections/31ce9ba0-176d-45a5-add3-f37d233ba47d
# replace with your desired collections
SRC_COLLECTION = "6c54cade-bde5-45c1-bdea-f4bd71dba2cc"
DST_COLLECTION = "31ce9ba0-176d-45a5-add3-f37d233ba47d"

SRC_PATH = "/home/share/godata/file1.txt"
DST_PATH = "/~/example-transfer-script-destination.txt"


with globus_sdk.UserApp(
    "detect-data-access-example", client_id=NATIVE_CLIENT_ID
) as app:
    with globus_sdk.TransferClient(app=app) as transfer_client:
        # check if either source or dest needs data_access, and if so add the relevant
        # requirement
        if uses_data_access(transfer_client, SRC_COLLECTION):
            transfer_client.add_app_data_access_scope(SRC_COLLECTION)
        if uses_data_access(transfer_client, DST_COLLECTION):
            transfer_client.add_app_data_access_scope(DST_COLLECTION)

        transfer_request = globus_sdk.TransferData(SRC_COLLECTION, DST_COLLECTION)
        transfer_request.add_item(SRC_PATH, DST_PATH)

        task = transfer_client.submit_transfer(transfer_request)
        print(f"Submitted transfer. Task ID: {task['task_id']}.")



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

project_id = str(uuid.uuid1())
star_lord = {
    "identity_provider": str(uuid.uuid1()),
    "identity_type": "login",
    "organization": "Guardians of the Galaxy",
    "status": "used",
    "id": str(uuid.uuid1()),
    "name": "Star Lord",
    "username": "star.lord@guardians.galaxy",
    "email": "star.lord2@guardians.galaxy",
}
guardians_group = {
    "id": str(uuid.uuid1()),
    "name": "Guardians of the Galaxy",
    "description": "A group of heroes",
    "organization": "Guardians of the Galaxy",
}


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        path=f"/v2/api/projects/{project_id}",
        method="PUT",
        json={
            "project": {
                "contact_email": "support@globus.org",
                "id": project_id,
                "admins": {
                    "identities": [star_lord],
                    "groups": [],
                },
                "project_name": "Guardians of the Galaxy",
                "admin_ids": [star_lord["id"]],
                "admin_group_ids": None,
                "display_name": "Guardians of the Galaxy",
            }
        },
        metadata={
            "id": project_id,
            "admin_id": star_lord["id"],
        },
    ),
    admin_group=RegisteredResponse(
        service="auth",
        path=f"/v2/api/projects/{project_id}",
        method="PUT",
        json={
            "project": {
                "contact_email": "support@globus.org",
                "id": project_id,
                "admins": {
                    "identities": [],
                    "groups": [guardians_group],
                },
                "project_name": "Guardians of the Galaxy",
                "admin_ids": None,
                "admin_group_ids": [guardians_group["id"]],
                "display_name": "Guardians of the Galaxy",
            }
        },
        metadata={
            "id": project_id,
            "admin_group_id": guardians_group["id"],
        },
    ),
)



import pytest

import globus_sdk


def test_create_job_rejects_transfer_timer():
    client = globus_sdk.TimersClient()
    payload = globus_sdk.TransferTimer(schedule={"type": "once"}, body={})

    with pytest.raises(
        globus_sdk.GlobusSDKUsageError,
        match=r"Cannot pass a TransferTimer to create_job\(\)\.",
    ):
        client.create_job(payload)


def test_create_timer_rejects_timer_job():
    client = globus_sdk.TimersClient()
    payload = globus_sdk.TimerJob("https://bogus", {}, "2021-01-01T00:00:00Z", 300)

    with pytest.raises(
        globus_sdk.GlobusSDKUsageError,
        match=r"Cannot pass a TimerJob to create_timer\(\)\.",
    ):
        client.create_timer(payload)



#!/usr/bin/env python

import argparse
import os
import sys

import globus_sdk
from globus_sdk.token_storage import SimpleJSONFileAdapter

MY_FILE_ADAPTER = SimpleJSONFileAdapter(os.path.expanduser("~/.sdk-manage-flow.json"))

# tutorial client ID
# we recommend replacing this with your own client for any production use-cases
CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"

NATIVE_CLIENT = globus_sdk.NativeAppAuthClient(CLIENT_ID)


def do_login_flow(scope):
    NATIVE_CLIENT.oauth2_start_flow(requested_scopes=scope, refresh_tokens=True)
    authorize_url = NATIVE_CLIENT.oauth2_get_authorize_url()
    print(f"Please go to this URL and login:\n\n{authorize_url}\n")
    auth_code = input("Please enter the code here: ").strip()
    tokens = NATIVE_CLIENT.oauth2_exchange_code_for_tokens(auth_code)
    return tokens


def get_authorizer(flow_id):
    scopes = globus_sdk.SpecificFlowClient(flow_id).scopes

    # try to load the tokens from the file, possibly returning None
    if MY_FILE_ADAPTER.file_exists():
        tokens = MY_FILE_ADAPTER.get_token_data(flow_id)
    else:
        tokens = None

    if tokens is None:
        # do a login flow, getting back initial tokens
        response = do_login_flow(scopes.user)
        # now store the tokens and pull out the correct token
        MY_FILE_ADAPTER.store(response)
        tokens = response.by_resource_server[flow_id]

    return globus_sdk.RefreshTokenAuthorizer(
        tokens["refresh_token"],
        NATIVE_CLIENT,
        access_token=tokens["access_token"],
        expires_at=tokens["expires_at_seconds"],
        on_refresh=MY_FILE_ADAPTER.on_refresh,
    )


def get_flow_client(flow_id):
    authorizer = get_authorizer(flow_id)
    return globus_sdk.SpecificFlowClient(flow_id, authorizer=authorizer)


def run_flow(args):
    flow_client = get_flow_client(args.FLOW_ID)
    print(flow_client.run_flow({}))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("FLOW_ID", help="Flow ID to run")
    args = parser.parse_args()

    try:
        run_flow(args)
    except globus_sdk.FlowsAPIError as e:
        print(f"API Error: {e.code} {e.message}")
        print(e.text)
        sys.exit(1)


if __name__ == "__main__":
    main()



import time
import uuid
from unittest import mock

import pytest

import globus_sdk
from globus_sdk.testing import RegisteredResponse
from globus_sdk.token_storage import TokenStorageData


@pytest.fixture
def id_token_sub():
    return str(uuid.UUID(int=1))


@pytest.fixture
def cc_auth_client():
    client = globus_sdk.ConfidentialAppAuthClient("dummy_id", "dummy_secret")
    with client.retry_config.tune(max_retries=0):
        yield client


@pytest.fixture
def mock_token_data_by_resource_server():
    expiration_time = int(time.time()) + 3600
    ret = {
        "resource_server_1": TokenStorageData(
            resource_server="resource_server_1",
            identity_id="user_id",
            scope="scope1",
            access_token="access_token_1",
            refresh_token="refresh_token_1",
            expires_at_seconds=expiration_time,
            token_type="Bearer",
        ),
        "resource_server_2": TokenStorageData(
            resource_server="resource_server_2",
            identity_id="user_id",
            scope="scope2 scope2:0 scope2:1",
            access_token="access_token_2",
            refresh_token="refresh_token_2",
            expires_at_seconds=expiration_time,
            token_type="Bearer",
        ),
    }
    return ret


@pytest.fixture
def mock_response():
    res = mock.Mock()
    expiration_time = int(time.time()) + 3600
    res.by_resource_server = {
        "resource_server_1": {
            "access_token": "access_token_1",
            "expires_at_seconds": expiration_time,
            "refresh_token": "refresh_token_1",
            "resource_server": "resource_server_1",
            "scope": "scope1",
            "token_type": "Bearer",
        },
        "resource_server_2": {
            "access_token": "access_token_2",
            "expires_at_seconds": expiration_time,
            "refresh_token": "refresh_token_2",
            "resource_server": "resource_server_2",
            "scope": "scope2 scope2:0 scope2:1",
            "token_type": "Bearer",
        },
    }
    res.decode_id_token.return_value = {"sub": "user_id"}

    return res


@pytest.fixture
def dependent_token_response(cc_auth_client):
    expiration_time = int(time.time()) + 3600
    RegisteredResponse(
        service="auth",
        path="/v2/oauth2/token",
        method="POST",
        json=[
            {
                "access_token": "access_token_1",
                "expires_at_seconds": expiration_time,
                "refresh_token": "refresh_token_1",
                "resource_server": "resource_server_1",
                "scope": "scope1",
                "token_type": "Bearer",
            },
            {
                "access_token": "access_token_2",
                "expires_at_seconds": expiration_time,
                "refresh_token": "refresh_token_2",
                "resource_server": "resource_server_2",
                "scope": "scope2 scope2:0 scope2:1",
                "token_type": "Bearer",
            },
        ],
    ).add()
    return cc_auth_client.oauth2_get_dependent_tokens("dummy_tok")


@pytest.fixture
def authorization_code_response(cc_auth_client, id_token_sub):
    cc_auth_client.oauth2_start_flow("https://example.com/redirect-uri", "dummy-scope")

    expiration_time = int(time.time()) + 3600
    RegisteredResponse(
        service="auth",
        path="/v2/oauth2/token",
        method="POST",
        json={
            "access_token": "access_token_1",
            "expires_at_seconds": expiration_time,
            "refresh_token": "refresh_token_1",
            "resource_server": "resource_server_1",
            "scope": "scope1",
            "token_type": "Bearer",
            "id_token": "dummy_id_token",
            "other_tokens": [
                {
                    "access_token": "access_token_2",
                    "expires_at_seconds": expiration_time,
                    "refresh_token": "refresh_token_2",
                    "resource_server": "resource_server_2",
                    "scope": "scope2 scope2:0 scope2:1",
                    "token_type": "Bearer",
                },
            ],
        },
    ).add()

    # because it's more difficult to mock the full decode_id_token() interaction in
    # detail, directly mock the result of it to return the desired subject (identity_id)
    # value
    response = cc_auth_client.oauth2_exchange_code_for_tokens("dummy_code")
    with mock.patch.object(response, "decode_id_token", lambda: {"sub": id_token_sub}):
        yield response


@pytest.fixture
def refresh_token_response(cc_auth_client):
    expiration_time = int(time.time()) + 3600
    RegisteredResponse(
        service="auth",
        path="/v2/oauth2/token",
        method="POST",
        json={
            "access_token": "access_token_1",
            "expires_at_seconds": expiration_time,
            "refresh_token": "refresh_token_1",
            "resource_server": "resource_server_1",
            "scope": "scope1",
            "token_type": "Bearer",
            "other_tokens": [],
        },
    ).add()
    return cc_auth_client.oauth2_refresh_token("dummy_token")



from responses.matchers import query_param_matcher

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from .._common import ENDPOINT_ID, ENDPOINT_ID_2, ENDPOINT_ID_3, NON_USER_ID, USER_ID

DEFAULT_RESPONSE_DOC = [
    {
        "uuid": ENDPOINT_ID,
        "name": "my-endpoint",
        "display_name": "My Endpoint",
        "owner": USER_ID,
    },
    {
        "uuid": ENDPOINT_ID_2,
        "name": "my-second-endpoint",
        "display_name": "My Second Endpoint",
        "owner": USER_ID,
    },
]

ANY_RESPONSE_DOC = [
    {
        "uuid": ENDPOINT_ID,
        "name": "my-endpoint",
        "display_name": "My Endpoint",
        "owner": USER_ID,
    },
    {
        "uuid": ENDPOINT_ID_2,
        "name": "my-second-endpoint",
        "display_name": "My Second Endpoint",
        "owner": USER_ID,
    },
    {
        "uuid": ENDPOINT_ID_3,
        "name": "public_endpoint",
        "display_name": "Public Endpoint",
        "owner": NON_USER_ID,
    },
]

RESPONSES = ResponseSet(
    metadata={
        "endpoint_id": ENDPOINT_ID,
        "endpoint_id_2": ENDPOINT_ID_2,
        "endpoint_id_3": ENDPOINT_ID_3,
    },
    default=RegisteredResponse(
        service="compute",
        path="/v2/endpoints",
        method="GET",
        json=DEFAULT_RESPONSE_DOC,
    ),
    any=RegisteredResponse(
        service="compute",
        path="/v2/endpoints",
        method="GET",
        json=ANY_RESPONSE_DOC,
        match=[query_param_matcher(params={"role": "any"})],
    ),
)



from globus_sdk.authorizers import NullAuthorizer


def test_get_authorization_header():
    """
    Gets authorization header. Confirms None value.
    """
    authorizer = NullAuthorizer()
    assert authorizer.get_authorization_header() is None


def test_handle_missing_authorization():
    """
    Confirms that NullAuthorizer doesn't handle missing authorization
    """
    authorizer = NullAuthorizer()
    assert not authorizer.handle_missing_authorization()



import random
import uuid

import pytest
import responses

from globus_sdk.paging import Paginator
from tests.common import register_api_route

# empty search
EMPTY_SEARCH_RESULT = {
    "DATA_TYPE": "endpoint_list",
    "offset": 0,
    "limit": 100,
    "has_next_page": False,
    "DATA": [],
}

# single page of data
SINGLE_PAGE_SEARCH_RESULT = {
    "DATA_TYPE": "endpoint_list",
    "offset": 0,
    "limit": 100,
    "has_next_page": False,
    "DATA": [
        {"DATA_TYPE": "endpoint", "display_name": f"SDK Test Stub {x}"}
        for x in range(100)
    ],
}

# multiple pages of results, very stubby
MULTIPAGE_SEARCH_RESULTS = [
    {
        "DATA_TYPE": "endpoint_list",
        "offset": 0,
        "limit": 100,
        "has_next_page": True,
        "DATA": [
            {"DATA_TYPE": "endpoint", "display_name": f"SDK Test Stub {x}"}
            for x in range(100)
        ],
    },
    {
        "DATA_TYPE": "endpoint_list",
        "offset": 100,
        "limit": 100,
        "has_next_page": True,
        "DATA": [
            {
                "DATA_TYPE": "endpoint",
                "display_name": f"SDK Test Stub {x + 100}",
            }
            for x in range(100, 200)
        ],
    },
    {
        "DATA_TYPE": "endpoint_list",
        "offset": 200,
        "limit": 100,
        "has_next_page": False,
        "DATA": [
            {
                "DATA_TYPE": "endpoint",
                "display_name": f"SDK Test Stub {x + 200}",
            }
            for x in range(100)
        ],
    },
]


def _mk_task_doc(idx):
    return {
        "DATA_TYPE": "task",
        "source_endpoint_id": "dc8e1110-b698-11eb-afd7-e1e7a67e00c1",
        "source_endpoint_display_name": "foreign place",
        "destination_endpoint_id": "83567b16-478d-4ead-a486-645bab0b07dc",
        "destination_endpoint_display_name": "my home",
        "directories": 0,
        "effective_bytes_per_second": random.randint(0, 10000),
        "files": 1,
        "encrypt_data": False,
        "label": f"autogen transfer {idx}",
    }


MULTIPAGE_OFFSET_TASK_LIST_RESULTS = [
    {
        "DATA_TYPE": "task_list",
        "offset": 0,
        "limit": 100,
        "total": 200,
        "DATA": [_mk_task_doc(x) for x in range(100)],
    },
    {
        "DATA_TYPE": "task_list",
        "offset": 100,
        "limit": 200,
        "total": 200,
        "DATA": [_mk_task_doc(x) for x in range(100, 200)],
    },
]


MULTIPAGE_LASTKEY_TASK_LIST_RESULTS = [
    {
        "DATA_TYPE": "task_list",
        "last_key": "abc",
        "limit": 100,
        "has_next_page": True,
        "DATA": [_mk_task_doc(x) for x in range(100)],
    },
    {
        "DATA_TYPE": "task_list",
        "last_key": "def",
        "limit": 100,
        "has_next_page": False,
        "DATA": [_mk_task_doc(x) for x in range(100, 200)],
    },
]


def test_endpoint_search_noresults(client):
    register_api_route("transfer", "/endpoint_search", json=EMPTY_SEARCH_RESULT)

    res = client.endpoint_search("search query!")
    assert res["DATA"] == []


def test_endpoint_search_one_page(client):
    register_api_route("transfer", "/endpoint_search", json=SINGLE_PAGE_SEARCH_RESULT)

    # without calling the paginated version, we only get one page
    res = client.endpoint_search("search query!")
    assert len(list(res)) == 100
    assert res["DATA_TYPE"] == "endpoint_list"
    for res_obj in res:
        assert res_obj["DATA_TYPE"] == "endpoint"


@pytest.mark.parametrize("method", ("__iter__", "pages"))
@pytest.mark.parametrize(
    "api_methodname,paged_data",
    [
        ("endpoint_search", MULTIPAGE_SEARCH_RESULTS),
        ("task_list", MULTIPAGE_OFFSET_TASK_LIST_RESULTS),
        ("endpoint_manager_task_list", MULTIPAGE_LASTKEY_TASK_LIST_RESULTS),
    ],
)
def test_paginated_method_multipage(client, method, api_methodname, paged_data):
    if api_methodname == "endpoint_search":
        route = "/endpoint_search"
        client_method = client.endpoint_search
        paginated_method = client.paginated.endpoint_search
        call_args = ("search_query",)
        wrapper_type = "endpoint_list"
        data_type = "endpoint"
    elif api_methodname == "task_list":
        route = "/task_list"
        client_method = client.task_list
        paginated_method = client.paginated.task_list
        call_args = ()
        wrapper_type = "task_list"
        data_type = "task"
    elif api_methodname == "endpoint_manager_task_list":
        route = "/endpoint_manager/task_list"
        client_method = client.endpoint_manager_task_list
        paginated_method = client.paginated.endpoint_manager_task_list
        call_args = ()
        wrapper_type = "task_list"
        data_type = "task"
    else:
        raise NotImplementedError

    # add each page
    for page in paged_data:
        register_api_route("transfer", route, json=page)

    # unpaginated, we'll only get one page
    res = list(client_method(*call_args))
    assert len(res) == 100

    # reset and reapply responses
    responses.reset()
    for page in paged_data:
        register_api_route("transfer", route, json=page)

    # setup the paginator and either point at `pages()` or directly at the paginator's
    # `__iter__`
    paginator = paginated_method(*call_args)
    if method == "pages":
        iterator = paginator.pages()
    elif method == "__iter__":
        iterator = paginator
    else:
        raise NotImplementedError

    # paginated calls gets all pages
    count_pages = 0
    count_objects = 0
    for page in iterator:
        count_pages += 1
        assert page["DATA_TYPE"] == wrapper_type
        for res_obj in page:
            count_objects += 1
            assert res_obj["DATA_TYPE"] == data_type

    assert count_pages == len(paged_data)
    assert count_objects == sum(len(x["DATA"]) for x in paged_data)


def test_endpoint_search_multipage_iter_items(client):
    # add each page
    for page in MULTIPAGE_SEARCH_RESULTS:
        register_api_route("transfer", "/endpoint_search", json=page)

    # paginator items() call gets an iterator of individual page items
    paginator = client.paginated.endpoint_search("search_query")
    count_objects = 0
    for item in paginator.items():
        count_objects += 1
        assert item["DATA_TYPE"] == "endpoint"

    assert count_objects == sum(len(x["DATA"]) for x in MULTIPAGE_SEARCH_RESULTS)


# multiple pages of results, very stubby
SHARED_ENDPOINT_RESULTS = [
    {
        "next_token": "token1",
        "shared_endpoints": [{"id": "abcd"} for x in range(1000)],
    },
    {
        "next_token": "token2",
        "shared_endpoints": [{"id": "abcd"} for x in range(1000)],
    },
    {
        "next_token": None,
        "shared_endpoints": [{"id": "abcd"} for x in range(100)],
    },
]


def test_shared_endpoint_list_non_paginated(client):
    # add each page
    for page in SHARED_ENDPOINT_RESULTS:
        register_api_route(
            "transfer", "/endpoint/endpoint_id/shared_endpoint_list", json=page
        )

    # without calling the paginated version, we only get one page
    res = client.get_shared_endpoint_list("endpoint_id")
    assert len(list(res)) == 1000
    for item in res:
        assert "id" in item


@pytest.mark.parametrize("paging_variant", ["attr", "wrap"])
def test_shared_endpoint_list_iter_pages(client, paging_variant):
    # add each page
    for page in SHARED_ENDPOINT_RESULTS:
        register_api_route(
            "transfer", "/endpoint/endpoint_id/shared_endpoint_list", json=page
        )

    # paginator pages() call gets an iterator of pages
    if paging_variant == "attr":
        paginator = client.paginated.get_shared_endpoint_list("endpoint_id")
    elif paging_variant == "wrap":
        paginator = Paginator.wrap(client.get_shared_endpoint_list)("endpoint_id")
    else:
        raise NotImplementedError
    count = 0
    for item in paginator.pages():
        count += 1
        assert "shared_endpoints" in item

    assert count == 3


@pytest.mark.parametrize("paging_variant", ["attr", "wrap"])
def test_shared_endpoint_list_iter_items(client, paging_variant):
    # add each page
    for page in SHARED_ENDPOINT_RESULTS:
        register_api_route(
            "transfer", "/endpoint/endpoint_id/shared_endpoint_list", json=page
        )

    # paginator items() call gets an iterator of individual page items
    if paging_variant == "attr":
        paginator = client.paginated.get_shared_endpoint_list("endpoint_id")
    elif paging_variant == "wrap":
        paginator = Paginator.wrap(client.get_shared_endpoint_list)("endpoint_id")
    else:
        raise NotImplementedError
    count = 0
    for item in paginator.items():
        count += 1
        assert "id" in item

    assert count == 2100


@pytest.mark.parametrize("paging_variant", ["attr", "wrap"])
def test_task_skipped_errors_pagination(client, paging_variant):
    task_id = str(uuid.uuid1())
    # add each page (10 pages)
    for page_number in range(10):
        page_data = []
        for item_number in range(100):
            page_data.append(
                {
                    "DATA_TYPE": "skipped_error",
                    "checksum_algorithm": None,
                    "destination_path": f"/~/{page_number}-{item_number}.txt",
                    "error_code": "PERMISSION_DENIED",
                    "error_details": "Error bad stuff happened",
                    "error_time": "2022-02-18T19:06:05+00:00",
                    "external_checksum": None,
                    "is_delete_destination_extra": False,
                    "is_directory": False,
                    "is_symlink": False,
                    "source_path": f"/~/{page_number}-{item_number}.txt",
                }
            )
        register_api_route(
            "transfer",
            f"/task/{task_id}/skipped_errors",
            json={
                "DATA_TYPE": "skipped_errors",
                "next_marker": f"mark{page_number}" if page_number < 9 else None,
                "DATA": page_data,
            },
        )

    # paginator items() call gets an iterator of individual page items
    if paging_variant == "attr":
        paginator = client.paginated.task_skipped_errors(task_id)
    elif paging_variant == "wrap":
        paginator = Paginator.wrap(client.task_skipped_errors)(task_id)
    else:
        raise NotImplementedError
    count = 0
    for item in paginator.items():
        count += 1
        assert item["DATA_TYPE"] == "skipped_error"

    assert count == 1000



from __future__ import annotations

import logging
import sys
import typing as t
import uuid

from globus_sdk import client, paging, response
from globus_sdk._internal.remarshal import commajoin, strseq_listify
from globus_sdk._missing import MISSING, MissingType
from globus_sdk.scopes import SearchScopes

from .data import SearchQueryV1, SearchScrollQuery
from .errors import SearchAPIError
from .response import IndexListResponse

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

log = logging.getLogger(__name__)

_VALID_ROLE_NAMES_T: TypeAlias = t.Literal["owner", "admin", "writer"]


class SearchClient(client.BaseClient):
    r"""
    Client for the Globus Search API

    .. sdk-sphinx-copy-params:: BaseClient

    This class provides helper methods for most common resources in the
    API, and basic ``get``, ``put``, ``post``, and ``delete`` methods
    from the base client that can be used to access any API resource.

    .. automethodlist:: globus_sdk.SearchClient
    """

    error_class = SearchAPIError
    service_name = "search"
    scopes = SearchScopes
    default_scope_requirements = [SearchScopes.search]

    #
    # Index Management
    #

    def create_index(
        self, display_name: str, description: str
    ) -> response.GlobusHTTPResponse:
        """
        Create a new index.

        :param display_name: the name of the index
        :param description: a description of the index

        New indices default to trial status. For subscribers with a subscription ID,
        indices can be converted to non-trial by sending a request to support@globus.org

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    r = sc.create_index(
                        "History and Witchcraft",
                        "Searchable information about history and witchcraft",
                    )
                    print(f"index ID: {r['id']}")

            .. tab-item:: Example Response Data

                .. expandtestfixture:: search.create_index

            .. tab-item:: API Info

                ``POST /v1/index``

                .. extdoclink:: Index Create
                    :ref: search/reference/index_create/
        """
        log.debug(f"SearchClient.create_index({display_name!r}, ...)")
        return self.post(
            "/v1/index", data={"display_name": display_name, "description": description}
        )

    def update_index(
        self,
        index_id: uuid.UUID | str,
        *,
        display_name: str | MissingType = MISSING,
        description: str | MissingType = MISSING,
    ) -> response.GlobusHTTPResponse:
        """
        Update index metadata.

        :param index_id: the ID of the index
        :param display_name: the name of the index
        :param description: a description of the index

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    MY_INDEX_ID = ...
                    r = sc.update_index(
                        MY_INDEX_ID,
                        display_name="My Awesome Index",
                        description="Very awesome searchable data",
                    )
                    print(f"index ID: {r['id']}")

            .. tab-item:: Example Response Data

                .. expandtestfixture:: search.create_index

            .. tab-item:: API Info

                ``PATCH /v1/index/<index_id>``

                .. extdoclink:: Index Update
                    :ref: search/reference/index_update/
        """
        log.debug(f"SearchClient.update_index({index_id!r}, ...)")
        return self.patch(
            f"/v1/index/{index_id}",
            data={"display_name": display_name, "description": description},
        )

    def delete_index(self, index_id: uuid.UUID | str) -> response.GlobusHTTPResponse:
        """
        Mark an index for deletion.

        Globus Search does not immediately delete indices. Instead, this API sets the
        index status to ``"delete-pending"``.
        Search will move pending tasks on the index to the ``CANCELLED`` state and will
        eventually delete the index.

        If the index is a trial index, it will be deleted a few minutes after being
        marked for deletion.
        If the index is non-trial, it will be kept for 30 days and will be eligible for
        use with the ``reopen`` API (see :meth:`~.reopen_index`) during that time.

        :param index_id: the ID of the index

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    sc.delete_index(index_id)

            .. tab-item:: Example Response Data

                .. expandtestfixture:: search.delete_index

            .. tab-item:: API Info

                ``DELETE /v1/index/<index_id>``

                .. extdoclink:: Index Delete
                    :ref: search/reference/index_delete/
        """
        log.debug(f"SearchClient.delete_index({index_id!r}, ...)")
        return self.delete(f"/v1/index/{index_id}")

    def reopen_index(self, index_id: uuid.UUID | str) -> response.GlobusHTTPResponse:
        """
        Reopen an index that has been marked for deletion, cancelling the deletion.

        :param index_id: the ID of the index

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    sc.reopen_index(index_id)

            .. tab-item:: Example Response Data

                .. expandtestfixture:: search.reopen_index

            .. tab-item:: API Info

                ``POST /v1/index/<index_id>/reopen``

                .. extdoclink:: Index Reopen
                    :ref: search/reference/index_reopen/
        """
        log.debug(f"SearchClient.reopen_index({index_id!r}, ...)")
        return self.post(f"/v1/index/{index_id}/reopen")

    def get_index(
        self,
        index_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Get descriptive data about a Search index, including its title and description
        and how much data it contains.

        :param index_id: the ID of the index
        :param query_params: additional parameters to pass as query params

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    index = sc.get_index(index_id)
                    assert index["id"] == index_id
                    print(index["display_name"], "(" + index_id + "):", index["description"])

            .. tab-item:: API Info

                ``GET /v1/index/<index_id>``

                .. extdoclink:: Index Show
                    :ref: search/reference/index_show/
        """  # noqa: E501
        log.debug(f"SearchClient.get_index({index_id})")
        return self.get(f"/v1/index/{index_id}", query_params=query_params)

    def index_list(
        self,
        filter_roles: (
            _VALID_ROLE_NAMES_T | t.Iterable[_VALID_ROLE_NAMES_T] | MissingType
        ) = MISSING,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.IterableResponse:
        """
        Get a list of indices on which the caller has permissions.

        :param filter_roles: An iterable of roles to use to filter the listing. By
            default, all indices where the user has a role are returned.
            Valid values are ``owner``, ``admin``, and ``writer``.
        :param query_params: additional parameters to pass as query params

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    for index_doc in sc.index_list():
                        print(index_doc["display_name"], f"({index_doc['id']}):")
                        print("  permissions:", ", ".join(index_doc["permissions"]))

            .. tab-item:: Example Response Data

                .. expandtestfixture:: search.index_list

            .. tab-item:: API Info

                ``GET /v1/index_list``

                .. extdoclink:: Index List
                    :ref: search/reference/index_list/
        """  # noqa: E501
        log.debug("SearchClient.index_list()")
        query_params = {"filter_roles": commajoin(filter_roles), **(query_params or {})}
        return IndexListResponse(self.get("/v1/index_list", query_params=query_params))

    #
    # Search queries
    #

    @paging.has_paginator(
        paging.HasNextPaginator,
        items_key="gmeta",
        get_page_size=lambda x: x["count"],
        max_total_results=10000,
        page_size=100,
    )
    def search(
        self,
        index_id: uuid.UUID | str,
        q: str,
        *,
        offset: int | MissingType = MISSING,
        limit: int | MissingType = MISSING,
        advanced: bool | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Execute a simple Search Query, described by the query string ``q``.

        :param index_id: the ID of the index
        :param q: the query string
        :param offset: an offset for pagination
        :param limit: the size of a page of results
        :param advanced: enable 'advanced' query mode, which has sophisticated syntax
            but may result in BadRequest errors when used if the query is invalid
        :param query_params: additional parameters to pass as query params

        For details on query syntax, including the ``advanced`` query behavior, see
        the :extdoclink:`Search Query Syntax <search/query#query_syntax>` documentation.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    result = sc.search(index_id, "query string")
                    advanced_result = sc.search(index_id, 'author: "Ada Lovelace"', advanced=True)

            .. tab-item:: Paginated Usage

                .. paginatedusage:: search

            .. tab-item:: API Info

                ``GET /v1/index/<index_id>/search``

                .. extdoclink:: GET Search Query
                    :ref: search/reference/get_query/

            .. tab-item:: Example Response Data

                .. expandtestfixture:: search.search
        """  # noqa: E501
        query_params = {
            "q": q,
            "offset": offset,
            "limit": limit,
            "advanced": advanced,
            **(query_params or {}),
        }
        log.debug(f"SearchClient.search({index_id}, ...)")
        return self.get(f"/v1/index/{index_id}/search", query_params=query_params)

    @paging.has_paginator(
        paging.HasNextPaginator,
        items_key="gmeta",
        get_page_size=lambda x: x["count"],
        max_total_results=10000,
        page_size=100,
    )
    def post_search(
        self,
        index_id: uuid.UUID | str,
        data: dict[str, t.Any] | SearchQueryV1,
        *,
        offset: int | MissingType = MISSING,
        limit: int | MissingType = MISSING,
    ) -> response.GlobusHTTPResponse:
        """
        Execute a complex Search Query, using a query document to express filters,
        facets, sorting, field boostring, and other behaviors.

        :param index_id: The index on which to search
        :param data: A Search Query document containing the query and any other fields
        :param offset: offset used in paging (overwrites any offset in ``data``)
        :param limit: limit the number of results (overwrites any limit in ``data``)

        For details on query syntax, including the ``advanced`` query behavior, see
        the :extdoclink:`Search Query Syntax <search/query#query_syntax>` documentation.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    query_data = {
                        "q": "user query",
                        "filters": [
                            {
                                "type": "range",
                                "field_name": "path.to.date",
                                "values": [{"from": "*", "to": "2014-11-07"}],
                            }
                        ],
                        "facets": [
                            {
                                "name": "Publication Date",
                                "field_name": "path.to.date",
                                "type": "date_histogram",
                                "date_interval": "year",
                            }
                        ],
                        "sort": [{"field_name": "path.to.date", "order": "asc"}],
                    }
                    search_result = sc.post_search(index_id, query_data)

            .. tab-item:: Paginated Usage

                .. paginatedusage:: post_search

            .. tab-item:: API Info

                ``POST /v1/index/<index_id>/search``

                .. extdoclink:: POST Search Query
                    :ref: search/reference/post_query/
        """
        log.debug(f"SearchClient.post_search({index_id}, ...)")
        add_kwargs = {}
        if offset is not MISSING:
            add_kwargs["offset"] = offset
        if limit is not MISSING:
            add_kwargs["limit"] = limit
        data = {**data, **add_kwargs}
        return self.post(f"v1/index/{index_id}/search", data=data)

    @paging.has_paginator(paging.MarkerPaginator, items_key="gmeta")
    def scroll(
        self,
        index_id: uuid.UUID | str,
        data: dict[str, t.Any] | SearchScrollQuery,
        *,
        marker: str | MissingType = MISSING,
    ) -> response.GlobusHTTPResponse:
        """
        Scroll all data in a Search index. The paginated version of this API should
        typically be preferred, as it is the intended mode of usage.

        Note that if data is written or deleted during scrolling, it is possible for
        scrolling to not include results or show other unexpected behaviors.

        :param index_id: The index on which to search
        :param data: A Search Scroll Query document
        :param marker: marker used in paging (overwrites any marker in ``data``)

        For details on query syntax, including the ``advanced`` query behavior, see
        the :extdoclink:`Search Query Syntax <search/query#query_syntax>` documentation.

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    scroll_result = sc.scroll(index_id, {"q": "*"})

            .. tab-item:: Paginated Usage

                .. paginatedusage:: scroll

            .. tab-item:: API Info

                ``POST /v1/index/<index_id>/scroll``

                .. extdoclink:: Scroll Query
                    :ref: search/reference/scroll_query/
        """
        log.debug(f"SearchClient.scroll({index_id}, ...)")
        add_kwargs = {}
        if marker is not MISSING:
            add_kwargs["marker"] = marker
        data = {**data, **add_kwargs}
        return self.post(f"v1/index/{index_id}/scroll", data=data)

    #
    # Bulk data indexing
    #

    def ingest(
        self, index_id: uuid.UUID | str, data: dict[str, t.Any]
    ) -> response.GlobusHTTPResponse:
        """
        Write data to a Search index as an asynchronous task.
        The data can be provided as a single document or list of documents, but only one
        ``task_id`` value will be included in the response.

        :param index_id: The index into which to write data
        :param data: an ingest document

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    ingest_data = {
                        "ingest_type": "GMetaEntry",
                        "ingest_data": {
                            "subject": "https://example.com/foo/bar",
                            "visible_to": ["public"],
                            "content": {"foo/bar": "some val"},
                        },
                    }
                    sc.ingest(index_id, ingest_data)

                or with multiple entries at once via a GMetaList:

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    ingest_data = {
                        "ingest_type": "GMetaList",
                        "ingest_data": {
                            "gmeta": [
                                {
                                    "subject": "https://example.com/foo/bar",
                                    "visible_to": ["public"],
                                    "content": {"foo/bar": "some val"},
                                },
                                {
                                    "subject": "https://example.com/foo/bar",
                                    "id": "otherentry",
                                    "visible_to": ["public"],
                                    "content": {"foo/bar": "some otherval"},
                                },
                            ]
                        },
                    }
                    sc.ingest(index_id, ingest_data)

            .. tab-item:: API Info

                ``POST /v1/index/<index_id>/ingest``

                .. extdoclink:: Ingest
                    :ref: search/reference/ingest/
        """
        log.debug(f"SearchClient.ingest({index_id}, ...)")
        return self.post(f"/v1/index/{index_id}/ingest", data=data)

    #
    # Bulk delete
    #

    def delete_by_query(
        self, index_id: uuid.UUID | str, data: dict[str, t.Any]
    ) -> response.GlobusHTTPResponse:
        """
        Delete data in a Search index as an asynchronous task, deleting all documents
        which match a given query.
        The query uses a restricted subset of the syntax available for complex queries,
        as it is not meaningful to boost, sort, or otherwise rank data in this case.

        A ``task_id`` value will be included in the response.

        :param index_id: The index in which to delete data
        :param data: a query document for documents to delete

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    query_data = {
                        "q": "user query",
                        "filters": [
                            {
                                "type": "range",
                                "field_name": "path.to.date",
                                "values": [{"from": "*", "to": "2014-11-07"}],
                            }
                        ],
                    }
                    sc.delete_by_query(index_id, query_data)

            .. tab-item:: API Info

                ``POST /v1/index/<index_id>/delete_by_query``

                .. extdoclink:: Delete By Query
                    :ref: search/reference/delete_by_query/
        """
        log.debug(f"SearchClient.delete_by_query({index_id}, ...)")
        return self.post(f"/v1/index/{index_id}/delete_by_query", data=data)

    def batch_delete_by_subject(
        self,
        index_id: uuid.UUID | str,
        subjects: t.Iterable[str],
        additional_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Delete data in a Search index as an asynchronous task, deleting multiple
        documents based on their ``subject`` values.

        A ``task_id`` value will be included in the response.

        :param index_id: The index in which to delete data
        :param subjects: The subjects to delete, as an iterable of strings
        :param additional_params: Additional parameters to include in the request body

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    sc.batch_delete_by_subject(
                        index_id,
                        subjects=[
                            "very-cool-document",
                            "less-cool-document",
                            "document-wearing-sunglasses",
                        ],
                    )

            .. tab-item:: Example Response Data

                .. expandtestfixture:: search.batch_delete_by_subject

            .. tab-item:: API Info

                ``POST /v1/index/<index_id>/batch_delete_by_subject``

                .. extdoclink:: Delete By Subject
                    :ref: search/reference/batch_delete_by_subject/
        """
        log.debug(f"SearchClient.batch_delete_by_subject({index_id}, ...)")
        # convert the provided subjects to a list and use the "safe iter" helper to
        # ensure that a single string is *not* treated as an iterable of strings,
        # which is usually not intentional
        body = {
            "subjects": strseq_listify(subjects),
            **(additional_params or {}),
        }
        return self.post(f"/v1/index/{index_id}/batch_delete_by_subject", data=body)

    #
    # Subject Operations
    #

    def get_subject(
        self,
        index_id: uuid.UUID | str,
        subject: str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Fetch exactly one Subject document from Search, containing one or more Entries.

        :param index_id: the index containing this Subject
        :param subject: the subject string to fetch
        :param query_params: additional parameters to pass as query params

        .. tab-set::

            .. tab-item:: Example Usage

                Fetch the data for subject ``http://example.com/abc`` from index
                ``index_id``:

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    subject_data = sc.get_subject(index_id, "http://example.com/abc")

            .. tab-item:: API Info

                ``GET /v1/index/<index_id>/subject``

                .. extdoclink:: Get By Subject
                    :ref: search/reference/get_subject/
        """
        log.debug(f"SearchClient.get_subject({index_id}, {subject}, ...)")
        query_params = {
            "subject": subject,
            **(query_params or {}),
        }
        return self.get(f"/v1/index/{index_id}/subject", query_params=query_params)

    def delete_subject(
        self,
        index_id: uuid.UUID | str,
        subject: str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Delete exactly one Subject document from Search, containing one or more Entries,
        as an asynchronous task.

        A ``task_id`` value will be included in the response.

        :param index_id: the index in which data will be deleted
        :param subject: the subject string for the Subject document to delete
        :param query_params: additional parameters to pass as query params

        .. tab-set::

            .. tab-item:: Example Usage

                Delete all data for subject ``http://example.com/abc`` from index
                ``index_id``, even data which is not visible to the current user:

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    response = sc.delete_subject(index_id, "http://example.com/abc")
                    task_id = response["task_id"]

            .. tab-item:: API Info

                ``DELETE /v1/index/<index_id>/subject``

                .. extdoclink:: Delete By Subject
                    :ref: search/reference/delete_subject/
        """
        log.debug(f"SearchClient.delete_subject({index_id}, {subject}, ...)")
        query_params = {
            "subject": subject,
            **(query_params or {}),
        }
        return self.delete(f"/v1/index/{index_id}/subject", query_params=query_params)

    #
    # Entry Operations
    #

    def get_entry(
        self,
        index_id: uuid.UUID | str,
        subject: str,
        *,
        entry_id: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Fetch exactly one Entry document from Search, identified by the combination of
        ``subject`` string and ``entry_id``, which defaults to ``null``.

        :param index_id: the index containing this Entry
        :param subject: the subject string for the Subject document containing this
            Entry
        :param entry_id: the entry_id for this Entry, which defaults to ``null``
        :param query_params: additional parameters to pass as query params

        .. tab-set::

            .. tab-item:: Example Usage

                Lookup the entry with a subject of ``https://example.com/foo/bar`` and
                a null entry_id:

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    entry_data = sc.get_entry(index_id, "http://example.com/foo/bar")

                Lookup the entry with a subject of ``https://example.com/foo/bar`` and
                an entry_id of ``foo/bar``:

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    entry_data = sc.get_entry(index_id, "http://example.com/foo/bar", entry_id="foo/bar")

            .. tab-item:: API Info

                ``GET /v1/index/<index_id>/entry``

                .. extdoclink:: Get Entry
                    :ref: search/reference/get_entry/
        """  # noqa: E501
        log.debug(
            "SearchClient.get_entry({}, {}, {}, ...)".format(
                index_id, subject, entry_id
            )
        )
        query_params = {
            "entry_id": entry_id,
            "subject": subject,
            **(query_params or {}),
        }
        return self.get(f"/v1/index/{index_id}/entry", query_params=query_params)

    def delete_entry(
        self,
        index_id: uuid.UUID | str,
        subject: str,
        *,
        entry_id: str | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Delete exactly one Entry document in Search as an asynchronous task.

        A ``task_id`` value will be included in the response.

        :param index_id: the index in which data will be deleted
        :param subject: the subject string for the Subject of the document to delete
        :param entry_id: the ID string for the Entry to delete
        :param query_params: additional parameters to pass as query params

        .. tab-set::

            .. tab-item:: Example Usage

                Delete an entry with a subject of ``https://example.com/foo/bar`` and
                a null entry_id:

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    sc.delete_entry(index_id, "https://example.com/foo/bar")

                Delete an entry with a subject of ``https://example.com/foo/bar`` and
                an entry_id of "foo/bar":

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    sc.delete_entry(index_id, "https://example.com/foo/bar", entry_id="foo/bar")

            .. tab-item:: API Info

                ``DELETE  /v1/index/<index_id>/entry``

                .. extdoclink:: Delete Entry
                    :ref: search/reference/delete_entry/
        """  # noqa: E501
        log.debug(
            "SearchClient.delete_entry({}, {}, {}, ...)".format(
                index_id, subject, entry_id
            )
        )
        query_params = {
            "entry_id": entry_id,
            "subject": subject,
            **(query_params or {}),
        }
        return self.delete(f"/v1/index/{index_id}/entry", query_params=query_params)

    #
    # Task Management
    #

    def get_task(
        self,
        task_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Fetch a Task document by ID, getting task details and status.

        :param task_id: the task ID from the original task submission
        :param query_params: additional parameters to pass as query params

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    task = sc.get_task(task_id)
                    assert task["index_id"] == known_index_id
                    print(task["task_id"], "|", task["state"])

            .. tab-item:: API Info

                ``GET /v1/task/<task_id>``

                .. extdoclink:: Get Task
                    :ref: search/reference/get_task/
        """
        log.debug(f"SearchClient.get_task({task_id})")
        return self.get(f"/v1/task/{task_id}", query_params=query_params)

    def get_task_list(
        self,
        index_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Fetch a list of recent Task documents for an index, getting task details and
        status.

        :param index_id: the index to query
        :param query_params: additional parameters to pass as query params

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    sc = globus_sdk.SearchClient(...)
                    task_list = sc.get_task_list(index_id)
                    for task in task_list["tasks"]:
                        print(task["task_id"], "|", task["state"])

            .. tab-item:: API Info

                ``GET /v1/task_list/<index_id>``

                .. extdoclink:: Task List
                    :ref: search/reference/task_list/
        """
        log.debug(f"SearchClient.get_task_list({index_id})")
        return self.get(f"/v1/task_list/{index_id}", query_params=query_params)

    #
    # Role Management
    #

    def create_role(
        self,
        index_id: uuid.UUID | str,
        data: dict[str, t.Any],
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Create a new role on an index. You must already have the ``owner`` or
        ``admin`` role on an index to create additional roles.

        Roles are specified as a role name (one of ``"owner"``, ``"admin"``, or
        ``"writer"``) and a `Principal URN
        <https://docs.globus.org/api/search/overview/#principal_urns>`_.

        :param index_id: The index on which to create the role
        :param data: The partial role document to use for creation
        :param query_params: Any additional query params to pass

        .. tab-set::

            .. tab-item:: Example Usage

                .. code-block:: python

                    identity_id = "46bd0f56-e24f-11e5-a510-131bef46955c"
                    sc = globus_sdk.SearchClient(...)
                    sc.create_role(
                        index_id,
                        {"role_name": "writer", "principal": f"urn:globus:auth:identity:{identity_id}"},
                    )

            .. tab-item:: API Info

                ``POST /v1/index/<index_id>/role``

                .. extdoclink:: Create Role
                    :ref: search/reference/role_create/
        """  # noqa: E501
        log.debug("SearchClient.create_role(%s, ...)", index_id)
        return self.post(
            f"/v1/index/{index_id}/role", data=data, query_params=query_params
        )

    def get_role_list(
        self,
        index_id: uuid.UUID | str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        List all roles on an index. You must have the ``owner`` or ``admin``
        role on an index to list roles.

        :param index_id: The index on which to list roles
        :param query_params: Any additional query params to pass

        .. tab-set::

            .. tab-item:: API Info

                ``GET /v1/index/<index_id>/role_list``

                .. extdoclink:: Get Role List
                    :ref: search/reference/role_list/
        """
        log.debug("SearchClient.get_role_list(%s)", index_id)
        return self.get(f"/v1/index/{index_id}/role_list", query_params=query_params)

    def delete_role(
        self,
        index_id: uuid.UUID | str,
        role_id: str,
        *,
        query_params: dict[str, t.Any] | None = None,
    ) -> response.GlobusHTTPResponse:
        """
        Delete a role from an index. You must have the ``owner`` or ``admin``
        role on an index to delete roles. You cannot remove the last ``owner`` from an
        index.

        :param index_id: The index from which to delete a role
        :param role_id: The role to delete
        :param query_params: Any additional query params to pass

        .. tab-set::

            .. tab-item:: API Info

                ``DELETE /v1/index/<index_id>/role/<role_id>``

                .. extdoclink:: Role Delete
                    :ref: search/reference/role_delete/
        """
        log.debug("SearchClient.delete_role(%s, %s)", index_id, role_id)
        return self.delete(
            f"/v1/index/{index_id}/role/{role_id}", query_params=query_params
        )



from __future__ import annotations

import http.client
import json
import typing as t

import requests
import responses

from globus_sdk.exc import GlobusAPIError

E = t.TypeVar("E", bound=GlobusAPIError)


def get_last_request(
    *, requests_mock: responses.RequestsMock | None = None
) -> requests.PreparedRequest | None:
    """
    Get the last request which was received, or None if there were no requests.

    :param requests_mock: A non-default ``RequestsMock`` object to use.
    """
    calls = requests_mock.calls if requests_mock is not None else responses.calls
    try:
        last_call = calls[-1]
    except IndexError:
        return None
    return last_call.request


@t.overload
def construct_error(
    *,
    http_status: int,
    body: bytes | str | t.Dict[str, t.Any],
    method: str = "GET",
    response_headers: t.Dict[str, str] | None = None,
    request_headers: t.Dict[str, str] | None = None,
    response_encoding: str = "utf-8",
    url: str = "https://bogus-url/",
) -> GlobusAPIError: ...


@t.overload
def construct_error(
    *,
    http_status: int,
    error_class: type[E],
    body: bytes | str | t.Dict[str, t.Any],
    method: str = "GET",
    response_headers: t.Dict[str, str] | None = None,
    request_headers: t.Dict[str, str] | None = None,
    response_encoding: str = "utf-8",
    url: str = "https://bogus-url/",
) -> E: ...


def construct_error(
    *,
    http_status: int,
    body: bytes | str | t.Dict[str, t.Any],
    error_class: type[E] | type[GlobusAPIError] = GlobusAPIError,
    method: str = "GET",
    response_headers: t.Dict[str, str] | None = None,
    request_headers: t.Dict[str, str] | None = None,
    response_encoding: str = "utf-8",
    url: str = "https://bogus-url/",
) -> E | GlobusAPIError:
    """
    Given parameters for an HTTP response, construct a GlobusAPIError and return it.

    :param error_class: The class of the error to construct. Defaults to
        GlobusAPIError.
    :param http_status: The HTTP status code to use in the response.
    :param body: The body of the response. If a dict, will be JSON-encoded.
    :param method: The HTTP method to set on the underlying request.
    :param response_headers: The headers of the response.
    :param request_headers: The headers of the request.
    :param response_encoding: The encoding to use for the response body.
    :param url: The URL to set on the underlying request.
    """
    raw_response = requests.Response()
    raw_response.status_code = http_status
    raw_response.reason = http.client.responses.get(http_status, "Unknown")
    raw_response.url = url
    raw_response.encoding = response_encoding
    raw_response.request = requests.Request(
        method=method, url=url, headers=request_headers or {}
    ).prepare()
    raw_response.headers.update(response_headers or {})
    if isinstance(body, dict) and "Content-Type" not in raw_response.headers:
        raw_response.headers["Content-Type"] = "application/json"

    raw_response._content = _encode_body(body, response_encoding)

    return error_class(raw_response)


def _encode_body(body: bytes | str | t.Dict[str, t.Any], encoding: str) -> bytes:
    if isinstance(body, bytes):
        return body
    elif isinstance(body, str):
        return body.encode(encoding)
    else:
        return json.dumps(body).encode(encoding)



from __future__ import annotations

import typing as t

from globus_sdk._internal.guards import validators
from globus_sdk._internal.serializable import Serializable


class TokenStorageData(Serializable):
    """
    Data class for tokens and token metadata issued by a globus auth token grant.
    For storage and retrieval of these objects, see :class:`TokenStorage`.

    Tokens are scoped to a specific user/client (`identity_id`) performing
    specific operations (`scope`) with a specific service (`resource_server`).

    :ivar str resource_server: The resource server for which this token data was
        granted.

    :ivar str identity_id: The primary identity id of the user or client which
        requested this token. This will be None if an identity id was not extractable
        from the token grant response.

    :ivar str scope: A space separated list of scopes that this token data provides
        access to.

    :ivar str access_token: A Globus Auth-issued OAuth2 access token. Used for
        authentication when interacting with service APIs.

    :ivar str | None refresh_token: A Globus Auth-issued OAuth2 refresh token. Used to
        obtain new access tokens when the current one expires. This value will be None
        if the original token grant did not request refresh tokens.

    :ivar int expires_at_seconds: An epoch seconds timestamp for when the associated
        access_token expires.

    :ivar str | None token_type: The token type of access_token, currently this will
        always be "Bearer" if present.

    :param extra: An optional dictionary of additional fields to include. Included for
        forward/backward compatibility.
    """

    def __init__(
        self,
        resource_server: str,
        identity_id: str | None,
        scope: str,
        access_token: str,
        refresh_token: str | None,
        expires_at_seconds: int,
        token_type: str | None,
        extra: dict[str, t.Any] | None = None,
    ) -> None:
        self.resource_server = validators.str_("resource_server", resource_server)
        self.identity_id = validators.opt_str("identity_id", identity_id)
        self.scope = validators.str_("scope", scope)
        self.access_token = validators.str_("access_token", access_token)
        self.refresh_token = validators.opt_str("refresh_token", refresh_token)
        self.expires_at_seconds = validators.int_(
            "expires_at_seconds", expires_at_seconds
        )
        self.token_type = validators.opt_str("token_type", token_type)
        self.extra = extra or {}



import urllib.parse
import uuid

import pytest

from globus_sdk import MISSING
from globus_sdk.testing import get_last_request, load_response


def test_list_runs_simple(flows_client):
    meta = load_response(flows_client.list_runs).metadata

    res = flows_client.list_runs()
    assert res.http_status == 200

    # dict-like indexing
    assert meta["first_run_id"] == res["runs"][0]["run_id"]
    # list conversion (using __iter__) and indexing
    assert meta["first_run_id"] == list(res)[0]["run_id"]


@pytest.mark.parametrize("by_pages", [True, False])
def test_list_runs_paginated(flows_client, by_pages):
    meta = load_response(flows_client.list_runs, case="paginated").metadata
    total_items = meta["total_items"]
    num_pages = meta["num_pages"]
    expect_markers = meta["expect_markers"]

    res = flows_client.paginated.list_runs()
    if by_pages:
        pages = list(res)
        assert len(pages) == num_pages
        for i, page in enumerate(pages):
            assert page["marker"] == expect_markers[i]
            if i < num_pages - 1:
                assert page["has_next_page"] is True
            else:
                assert page["has_next_page"] is False
    else:
        items = list(res.items())
        assert len(items) == total_items


@pytest.mark.parametrize("pass_as_uuids", [True, False])
def test_list_runs_filter_flow_id(flows_client, pass_as_uuids):
    meta = load_response(flows_client.list_runs, case="filter_flow_id").metadata
    # sanity check that the underlying test data hasn't changed too much
    assert len(meta["by_flow_id"]) == 2
    flow_id_one, flow_id_two = tuple(meta["by_flow_id"].keys())
    if pass_as_uuids:
        flow_id_one = uuid.UUID(flow_id_one)
        flow_id_two = uuid.UUID(flow_id_two)

    res_one = list(flows_client.list_runs(filter_flow_id=flow_id_one))
    assert len(res_one) == meta["by_flow_id"][str(flow_id_one)]["num"]
    for run in res_one:
        assert run["flow_id"] == str(flow_id_one)

    res_two = list(flows_client.list_runs(filter_flow_id=flow_id_two))
    assert len(res_two) == meta["by_flow_id"][str(flow_id_two)]["num"]
    for run in res_two:
        assert run["flow_id"] == str(flow_id_two)

    res_combined = list(
        flows_client.list_runs(filter_flow_id=[flow_id_one, flow_id_two])
    )
    assert len(res_combined) == (
        meta["by_flow_id"][str(flow_id_one)]["num"]
        + meta["by_flow_id"][str(flow_id_two)]["num"]
    )
    for run in res_combined:
        assert run["flow_id"] in {str(flow_id_one), str(flow_id_two)}


@pytest.mark.parametrize(
    "filter_roles, expected_filter_roles",
    [
        # empty list/tuple, list/tuple with empty string, and None do not send the param
        ([], None),
        ((), None),
        ([""], None),
        (("",), None),
        (MISSING, None),
        # single role as string
        ("foo", ["foo"]),
        # single role as list/tuple
        (["foo"], ["foo"]),
        (("foo",), ["foo"]),
        # multiple roles as comma-separated string
        ("foo,bar", ["foo,bar"]),
        # multiple roles as list/tuple
        (["foo", "bar"], ["foo,bar"]),
        (("foo", "bar"), ["foo,bar"]),
    ],
)
def test_list_runs_filter_roles(flows_client, filter_roles, expected_filter_roles):
    load_response(flows_client.list_runs).metadata
    res = flows_client.list_runs(filter_roles=filter_roles)
    assert res.http_status == 200

    req = get_last_request()
    assert req.body is None
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)

    if expected_filter_roles:
        assert parsed_qs["filter_roles"] == expected_filter_roles
    else:
        assert "filter_roles" not in parsed_qs



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

CLIENT = {
    "required_idp": None,
    "name": "Great client of FOO",
    "redirect_uris": [],
    "links": {"privacy_policy": None, "terms_and_conditions": None},
    "scopes": [],
    "grant_types": ["authorization_code", "client_credentials", "refresh_token"],
    "id": str(uuid.uuid1()),
    "prompt_for_named_grant": False,
    "fqdns": ["globus.org"],
    "project": "da84e531-1afb-43cb-8c87-135ab580516a",
    "client_type": "client_identity",
    "visibility": "private",
    "parent_client": None,
    "userinfo_from_effective_identity": True,
    "preselect_idp": None,
    "public_client": False,
}

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        method="DELETE",
        path=f"/v2/api/clients/{CLIENT['id']}",
        json={"client": CLIENT},
        metadata={
            "client_id": CLIENT["id"],
        },
    ),
)



from __future__ import annotations

import json
import typing as t

from globus_sdk import __version__

from .base import FileTokenStorage
from .token_data import TokenStorageData

# use the non-annotation form of TypedDict to apply a non-identifier key
_JSONFileData_0 = t.TypedDict("_JSONFileData_0", {"globus-sdk.version": str})


# then inherit from that TypedDict to build the "real" TypedDict with the advantages of
# the annotation-based syntax
class _JSONFileData(_JSONFileData_0):  # pylint: disable=inherit-non-class
    data: dict[str, dict[str, t.Any]]
    format_version: str


class JSONTokenStorage(FileTokenStorage):
    """
    A token storage which stores token data on disk in a JSON file.

    This class defines a `format_version` which determines what the specific data shape.
    Any data in a `supported_version` format which is not the primary `format_version`
    will be automatically rewritten.

    See :class:`TokenStorage` for common interface details.

    :cvar "2.0" format_version: The data format version used when writing data.
    :cvar ("1.0", "2.0") supported_versions: The list of data format versions which can
        be read.
    :param filepath: The path to a JSON file where token data should be stored.
    :param namespace: A unique string for partitioning token data (Default: "DEFAULT").
    """

    format_version = "2.0"
    supported_versions = ("1.0", "2.0")
    file_format = "json"

    def _invalid(self, msg: str) -> t.NoReturn:
        raise ValueError(
            f"{msg} while loading from '{self.filepath}' for JSON Token Storage"
        )

    def _raw_load(self) -> dict[str, t.Any]:
        """
        Load the file contents as JSON and return the resulting dict
        object. If a dict is not found, raises an error.
        """
        with open(self.filepath, encoding="utf-8") as f:
            val = json.load(f)
        if not isinstance(val, dict):
            self._invalid("Found non-dict root data")
        return val

    def _handle_formats(self, read_data: dict[str, t.Any]) -> _JSONFileData:
        """Handle older data formats supported by this class

        if the data is not in a known/recognized format, this will error
        otherwise, reshape the data to the current supported format and return it
        """
        format_version = read_data.get("format_version")
        if format_version not in self.supported_versions:
            raise ValueError(
                f"cannot store data using SimpleJSONTokenStorage({self.filepath}) "
                "existing data file is in an unknown format "
                f"(format_version={format_version})"
            )

        # 1.0 data was stored under a "by_rs" key without namespaces, to upgrade we
        # move everything under the "DEFAULT" key and remove the "by_rs" key.
        if format_version == "1.0":
            if "by_rs" not in read_data:
                self._invalid("Invalid v1.0 data (missing 'by_rs')")
            read_data = {
                "data": {
                    "DEFAULT": read_data["by_rs"],
                },
                "format_version": self.format_version,
                "globus-sdk.version": __version__,
            }

        if not isinstance(read_data.get("data"), dict):
            raise ValueError(
                f"cannot store data using SimpleJSONTokenStorage({self.filepath}) "
                "existing data file is malformed"
            )
        if any(
            k not in read_data for k in ("data", "format_version", "globus-sdk.version")
        ):
            self._invalid("Missing required keys")
        if not isinstance(data_dict := read_data["data"], dict) or any(
            not isinstance(k, str) for k in data_dict
        ):
            self._invalid("Invalid 'data'")
        if not isinstance(read_data["format_version"], str):
            self._invalid("Invalid 'format_version'")
        if not isinstance(read_data["globus-sdk.version"], str):
            self._invalid("Invalid 'globus-sdk.version'")

        return read_data  # type: ignore[return-value]

    def _load(self) -> _JSONFileData:
        """
        Load data from the file and ensure that the data is in a modern format which can
        be handled by the rest of the adapter.

        If the file is missing, this will return a "skeleton" for new data.
        """
        try:
            data = self._raw_load()
        except FileNotFoundError:
            return {
                "data": {},
                "format_version": self.format_version,
                "globus-sdk.version": __version__,
            }
        return self._handle_formats(data)

    def store_token_data_by_resource_server(
        self, token_data_by_resource_server: t.Mapping[str, TokenStorageData]
    ) -> None:
        """
        Store token data for one or more resource server in the current namespace.

        Token data, alongside Globus SDK version info, is serialized to JSON before
        being written to the file at ``self.filepath``.

        Under the assumption that this may be running on a system with multiple
        local users, this sets the umask such that only the owner of the
        resulting file can read or write it.

        :param token_data_by_resource_server: A mapping of resource servers to token
            data.
        """
        to_write = self._load()

        # create the namespace if it does not exist
        if self.namespace not in to_write["data"]:
            to_write["data"][self.namespace] = {}

        # add token data by resource server to namespaced data
        for resource_server, token_data in token_data_by_resource_server.items():
            to_write["data"][self.namespace][resource_server] = token_data.to_dict()

        # update globus-sdk version
        to_write["globus-sdk.version"] = __version__

        # write the file, denying rwx to Group and World, exec to User
        with self.user_only_umask():
            with open(self.filepath, "w", encoding="utf-8") as f:
                json.dump(to_write, f)

    def get_token_data_by_resource_server(self) -> dict[str, TokenStorageData]:
        """
        Lookup all token data under the current namespace from the JSON file.

        Returns a dict of ``TokenStorageData`` objects indexed by their resource server.
        """
        ret = {}
        dicts_by_resource_server = self._load()["data"].get(self.namespace, {})
        for resource_server, token_data_dict in dicts_by_resource_server.items():
            ret[resource_server] = TokenStorageData.from_dict(token_data_dict)
        return ret

    def remove_token_data(self, resource_server: str) -> bool:
        """
        Remove all tokens for a resource server from the JSON data, then overwrite
        ``self.filepath``.

        Returns True if token data was removed, False if none was found to remove.

        :param resource_server: The resource server string to remove tokens for
        """
        to_write = self._load()

        # pop the token data out if it exists
        popped = to_write["data"].get(self.namespace, {}).pop(resource_server, None)

        # overwrite the file, denying rwx to Group and World, exec to User
        with self.user_only_umask():
            with open(self.filepath, "w", encoding="utf-8") as f:
                json.dump(to_write, f)

        return popped is not None



from .base import FileTokenStorage, TokenStorage
from .json import JSONTokenStorage
from .memory import MemoryTokenStorage
from .sqlite import SQLiteTokenStorage
from .token_data import TokenStorageData
from .validating_token_storage import (
    HasRefreshTokensValidator,
    NotExpiredValidator,
    ScopeRequirementsValidator,
    TokenDataValidator,
    TokenValidationContext,
    TokenValidationError,
    UnchangingIdentityIDValidator,
    ValidatingTokenStorage,
)

__all__ = (
    "TokenStorage",
    "TokenStorageData",
    "FileTokenStorage",
    "JSONTokenStorage",
    "SQLiteTokenStorage",
    "MemoryTokenStorage",
    # TokenValidationStorage constructs
    "ValidatingTokenStorage",
    "TokenValidationContext",
    "TokenDataValidator",
    "TokenValidationError",
    "HasRefreshTokensValidator",
    "NotExpiredValidator",
    "ScopeRequirementsValidator",
    "UnchangingIdentityIDValidator",
)



# test typing interactions with __getattr__-based lazy imports
import globus_sdk

# ensure that a valid name is treated as valid
globus_sdk.TransferClient
# but an invalid name is not!
globus_sdk.TRansferClient  # type: ignore[attr-defined]
